"""Retrieve metadata from an FMR instance."""

from enum import Enum
from typing import (
    Any,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from msgspec.json import decode

from pysdmx.api.qb import (
    ApiVersion,
    AsyncRestService,
    RefMetaByMetadatasetQuery,
    RefMetaByStructureQuery,
    RestService,
    SchemaContext,
    SchemaQuery,
    StructureDetail,
    StructureQuery,
    StructureReference,
    StructureType,
)
from pysdmx.errors import NotFound, NotImplemented
from pysdmx.io.format import RefMetaFormat, SchemaFormat, StructureFormat
from pysdmx.io.json.fusion.reader import deserializers as fusion_readers
from pysdmx.io.json.sdmxjson2.reader import deserializers as sdmx_readers
from pysdmx.io.serde import Deserializer
from pysdmx.model import (
    Agency,
    Categorisation,
    CategoryScheme,
    Codelist,
    ConceptScheme,
    Dataflow,
    DataflowInfo,
    DataProvider,
    Hierarchy,
    HierarchyAssociation,
    MetadataReport,
    MultiRepresentationMap,
    ProvisionAgreement,
    RepresentationMap,
    Schema,
    StructureMap,
)


class DataflowDetails(Enum):
    """The list of allowed details for dataflow queries."""

    ALL = "all"
    """All available information about a dataflow."""
    CORE = "core"
    """Core information about a dataflow (ID, name, description, etc.)."""
    PROVIDERS = "providers"
    """Core information about a dataflow and the list of providers."""
    SCHEMA = "schema"
    """Core information about a dataflow and its schema (data structure.)"""


API_VERSION = ApiVersion.V2_0_0


class __BaseRegistryClient:
    __schema_q = [DataflowDetails.ALL, DataflowDetails.SCHEMA]
    __prov_q = [DataflowDetails.ALL, DataflowDetails.PROVIDERS]

    def __init__(
        self,
        api_endpoint: str,
        fmt: StructureFormat = StructureFormat.SDMX_JSON_2_0_0,
    ):
        """Instantiate a new client against the target endpoint."""
        if api_endpoint.endswith("/"):
            api_endpoint = api_endpoint[0:-1]
        self.api_endpoint = api_endpoint
        if fmt not in (
            StructureFormat.SDMX_JSON_2_0_0,
            StructureFormat.FUSION_JSON,
        ):
            raise NotImplemented(
                "Unsupported format",
                "Only SDMX-JSON v2.0.0 and Fusion-JSON are supported.",
                {"requested_format": fmt.value},
            )
        self.format = fmt
        if fmt == StructureFormat.FUSION_JSON:
            self.deser = fusion_readers
        else:
            self.deser = sdmx_readers

    def _out(self, response: bytes, typ: Deserializer, *params: Any) -> Any:
        return decode(response, type=typ).to_model(*params)

    def _df_details(
        self, details: DataflowDetails
    ) -> Tuple[bool, StructureReference]:
        sq = False
        dr = StructureReference.NONE
        if details in self.__schema_q:
            sq = True
        if details in self.__prov_q:
            dr = StructureReference.PARENTSANDSIBLINGS
        return (sq, dr)

    def _hierarchies_for_flow_q(
        self, agency: str, flow: str, version: str
    ) -> StructureQuery:
        d = StructureDetail.REFERENCE_PARTIAL
        r = StructureReference.ALL
        return StructureQuery(
            StructureType.DATAFLOW,
            agency,
            flow,
            version,
            detail=d,
            references=r,
        )

    def _hierarchies_for_pra_q(
        self, agency: str, pra: str, version: str
    ) -> StructureQuery:
        d = StructureDetail.REFERENCE_PARTIAL
        r = StructureReference.ALL
        return StructureQuery(
            StructureType.PROVISION_AGREEMENT,
            agency,
            pra,
            version,
            detail=d,
            references=r,
        )

    def _code_map_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(
            StructureType.REPRESENTATION_MAP, agency, id, version
        )

    def _mapping_q(self, agency: str, id: str, version: str) -> StructureQuery:
        r = StructureReference.CHILDREN
        d = StructureDetail.REFERENCE_PARTIAL
        return StructureQuery(
            StructureType.STRUCTURE_MAP,
            agency,
            id,
            version,
            detail=d,
            references=r,
        )

    def _reports_q(
        self,
        artefact_type: str,
        agency: str,
        id: str,
        version: str,
    ) -> RefMetaByStructureQuery:
        return RefMetaByStructureQuery(
            StructureType(artefact_type), agency, id, version
        )

    def _report_q(
        self, provider: str, id: str, version: str
    ) -> RefMetaByMetadatasetQuery:
        return RefMetaByMetadatasetQuery(provider, id, version)

    def _hierarchy_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(
            StructureType.HIERARCHY,
            agency,
            id,
            version,
            detail=StructureDetail.REFERENCE_PARTIAL,
            references=StructureReference.CODELIST,
        )

    def _agencies_q(self, agency: str) -> StructureQuery:
        return StructureQuery(StructureType.AGENCY_SCHEME, agency)

    def _providers_q(self, agency: str, with_flows: bool) -> StructureQuery:
        r = (
            StructureReference.PROVISION_AGREEMENT
            if with_flows
            else StructureReference.NONE
        )
        return StructureQuery(
            StructureType.DATA_PROVIDER_SCHEME, agency, references=r
        )

    def _categories_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(
            StructureType.CATEGORY_SCHEME,
            agency,
            id,
            version,
            detail=StructureDetail.REFERENCE_PARTIAL,
            references=StructureReference.PARENTSANDSIBLINGS,
        )

    def _codes_cl_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(StructureType.CODELIST, agency, id, version)

    def _codes_vl_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(StructureType.VALUE_LIST, agency, id, version)

    def _concepts_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(
            StructureType.CONCEPT_SCHEME,
            agency,
            id,
            version,
            references=StructureReference.CODELIST,
        )

    def _schema_q(
        self, context: SchemaContext, agency: str, id: str, version: str
    ) -> SchemaQuery:
        return SchemaQuery(context, agency, id, version)

    def _dataflow_details_q(
        self, agency: str, id: str, version: str, ref: StructureReference
    ) -> StructureQuery:
        return StructureQuery(
            StructureType.DATAFLOW,
            agency,
            id,
            version,
            detail=StructureDetail.REFERENCE_PARTIAL,
            references=ref,
        )

    def _dataflows_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(StructureType.DATAFLOW, agency, id, version)

    def _vtl_ts_q(self, agency: str, id: str, version: str) -> StructureQuery:
        return StructureQuery(
            StructureType.TRANSFORMATION_SCHEME,
            agency,
            id,
            version,
            detail=StructureDetail.REFERENCE_PARTIAL,
            references=StructureReference.DESCENDANTS,
        )

    def _categorisation_q(
        self, agency: str, id: str, version: str
    ) -> StructureQuery:
        return StructureQuery(
            StructureType.CATEGORISATION, agency, id, version
        )

    def _pa_q(self, agency: str, id: str, version: str) -> StructureQuery:
        return StructureQuery(
            StructureType.PROVISION_AGREEMENT, agency, id, version
        )


class RegistryClient(__BaseRegistryClient):
    """A client to be used to retrieve metadata from the FMR.

    With this client, metadata will be retrieved in a synchronous fashion.
    """

    def __init__(
        self,
        api_endpoint: str,
        format: StructureFormat = StructureFormat.SDMX_JSON_2_0_0,
        pem: Optional[str] = None,
        timeout: float = 10.0,
    ):
        """Instantiate a new client against the target endpoint.

        Args:
            api_endpoint: The endpoint of the targeted service.
            format: The format the service should use to serialize
                the metadata to be returned. Defaults to SDMX-JSON.
            pem: In case the service exposed a certificate created
                by an unknown certificate authority, you can pass
                a pem file for this authority using this parameter.
            timeout: The maximum number of seconds to wait before
                considering that a request timed out. Defaults to
                10 seconds.
        """
        super().__init__(api_endpoint, format)
        self.__service = RestService(
            self.api_endpoint,
            API_VERSION,
            structure_format=format,
            schema_format=(
                SchemaFormat.SDMX_JSON_2_0_0_STRUCTURE
                if format == StructureFormat.SDMX_JSON_2_0_0
                else SchemaFormat.FUSION_JSON
            ),
            refmeta_format=(
                RefMetaFormat.SDMX_JSON_2_0_0
                if format == StructureFormat.SDMX_JSON_2_0_0
                else RefMetaFormat.FUSION_JSON
            ),
            pem=pem,
            timeout=timeout,
        )

    def __fetch(
        self,
        query: Union[
            RefMetaByMetadatasetQuery,
            RefMetaByStructureQuery,
            SchemaQuery,
            StructureQuery,
        ],
    ) -> bytes:
        if isinstance(query, StructureQuery):
            return self.__service.structure(query)
        elif isinstance(query, SchemaQuery):
            return self.__service.schema(query)
        else:
            return self.__service.reference_metadata(query)

    def __get_hierarchies_for_flow(
        self, agency: str, flow: str, version: str
    ) -> Sequence[HierarchyAssociation]:
        query = super()._hierarchies_for_flow_q(agency, flow, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.hier_assoc)

    def __get_hierarchies_for_pra(
        self, agency: str, pra: str, version: str
    ) -> Sequence[HierarchyAssociation]:
        query = super()._hierarchies_for_pra_q(agency, pra, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.hier_assoc)

    def get_agencies(self, agency: str) -> Sequence[Agency]:
        """Get the list of **sub-agencies** for the supplied agency.

        Args:
            agency: The agency maintaining the agency scheme from
                which sub-agencies must be returned.

        Returns:
            The requested list of agencies.
        """
        query = super()._agencies_q(agency)
        out = self.__fetch(query)
        schemes = super()._out(out, self.deser.agencies)
        return schemes[0].items

    def get_providers(
        self,
        agency: str,
        with_flows: bool = False,
    ) -> Sequence[DataProvider]:
        """Get the list of **data providers** for the supplied agency.

        Args:
            agency: The agency maintaining the data provider scheme from
                which data providers must be returned.
            with_flows: Whether the data providers should contain the list
                of dataflows for which the data provider provides data.

        Returns:
            The requested list of data providers.
        """
        query = super()._providers_q(agency, with_flows)
        out = self.__fetch(query)
        schemes = super()._out(out, self.deser.providers)
        return schemes[0].items

    def get_categories(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> CategoryScheme:
        """Get the category scheme matching the supplied parameters.

        Args:
            agency: The agency maintaining the category scheme.
            id: The ID of the category scheme to be returned.
            version: The version of the category scheme to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested category scheme.
        """
        query = super()._categories_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.categories)

    def get_categorisation(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Categorisation:
        """Get the categorisation matching the supplied parameters.

        Args:
            agency: The agency maintaining the categorisation.
            id: The ID of the categorisation to be returned.
            version: The version of the categorisation to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested categorisation.
        """
        query = super()._categorisation_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.categorisation)[0]

    def get_provision_agreement(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> ProvisionAgreement:
        """Get the provision agreement matching the supplied parameters.

        Args:
            agency: The agency maintaining the provision agreement.
            id: The ID of the provision agreement to be returned.
            version: The version of the provision agreement to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested provision agreement.
        """
        query = super()._pa_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.provision_agreement)[0]

    def get_codes(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Codelist:
        """Get the codelist or valuelist matching the supplied parameters.

        Args:
            agency: The agency maintaining the codelist.
            id: The ID of the codelist to be returned.
            version: The version of the codelist to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested codelist.
        """
        try:
            query = super()._codes_cl_q(agency, id, version)
            out = self.__fetch(query)
            return super()._out(out, self.deser.codes)
        except NotFound:
            query = super()._codes_vl_q(agency, id, version)
            out = self.__fetch(query)
            return super()._out(out, self.deser.codes)

    def get_concepts(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> ConceptScheme:
        """Get the concept scheme matching the supplied parameters.

        Args:
            agency: The agency maintaining the concept scheme.
            id: The ID of the concept scheme to be returned.
            version: The version of the concept scheme to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested concept scheme.
        """
        query = super()._concepts_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.concepts)

    def get_schema(
        self,
        context: Union[
            SchemaContext,
            Literal["dataflow", "datastructure", "provisionagreement"],
        ],
        agency: str,
        id: str,
        version: str,
    ) -> Schema:
        """Get the schema matching the supplied parameters.

        Args:
            context: The context for which the schema should be generated
                (dataflow, datastructure or provisionagreement). This defines
                the constraints to be applied when generating the schema.
            agency: The ID of the agency maintaining the context.
            id: The ID of the context to be considered.
            version: The version of the context to be considered.

        Returns:
            The requested schema.
        """
        c = (
            context
            if isinstance(context, SchemaContext)
            else SchemaContext(context)
        )
        if c == SchemaContext.DATAFLOW:
            ha = self.__get_hierarchies_for_flow(agency, id, version)
        elif c == SchemaContext.PROVISION_AGREEMENT:
            ha = self.__get_hierarchies_for_pra(agency, id, version)
        else:
            ha = ()
        query = super()._schema_q(c, agency, id, version)
        out = self.__fetch(query)
        return super()._out(
            out, self.deser.schema, c.value, agency, id, version, ha
        )

    def get_dataflow_details(
        self,
        agency: str,
        id: str,
        version: str = "+",
        detail: Union[
            DataflowDetails, Literal["all", "core", "providers", "schema"]
        ] = DataflowDetails.ALL,
    ) -> DataflowInfo:
        """Get detailed information about a dataflow.

        Args:
            agency: The agency maintaining the dataflow.
            id: The ID of the dataflow to be returned.
            version: The version of the dataflow to be returned. The most
                recent version will be returned, unless specified otherwise.
            detail: The amount of detail to be returned. "core" means only
                core information (ID, name, etc.) will be returned.
                "providers" means core information will be returned, as well
                as the list of organizations providing data for the dataflow.
                "schema" means core information will be returned, as well
                as the schema describing the expected/allowed structure of
                data to be reported against the dataflow. "all" combines all
                3 other options.

        Returns:
            The requested information about a dataflow.
        """
        d = DataflowDetails(detail) if isinstance(detail, str) else detail
        sq, dr = super()._df_details(d)
        if sq:
            cmps = self.get_schema("dataflow", agency, id, version).components
        else:
            cmps = None
        query = super()._dataflow_details_q(agency, id, version, dr)
        out = self.__fetch(query)
        return super()._out(
            out, self.deser.dataflow_info, cmps, agency, id, version
        )

    def get_dataflows(
        self,
        agency: str = "*",
        id: str = "*",
        version: str = "+",
    ) -> Sequence[Dataflow]:
        """Get the dataflow(s) matching the supplied parameters.

        Args:
            agency: The agency maintaining the dataflow(s).
            id: The ID of the dataflow(s) to be returned.
            version: The version of the dataflow(s) to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested dataflow(s).
        """
        query = super()._dataflows_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.dataflows)

    def get_hierarchy(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Hierarchy:
        """Get a hierarchical list of codes.

        Args:
            agency: The agency maintaining the hierarchy.
            id: The ID of the hierarchy to be returned.
            version: The version of the hierarchy to be returned. The most
                recent version will be returned, unless specified otherwise.

        Returns:
            The requested hierarchical list of codes.
        """
        query = super()._hierarchy_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.hierarchy)

    def get_report(
        self,
        provider: str,
        id: str,
        version: str = "+",
    ) -> MetadataReport:
        """Get a reference metadata report.

        Args:
            provider: The organization which provided the report.
            id: The ID of the report to be returned.
            version: The version of the report to be returned. The most
                recent version will be returned, unless specified otherwise.

        Returns:
            The requested metadata report.
        """
        query = super()._report_q(provider, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.report)[0]

    def get_reports(
        self,
        artefact_type: str,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Sequence[MetadataReport]:
        """Get the reference metadata reports for the supplied structure.

        Args:
            artefact_type: The type of the structure for which reports must
                be returned.
            agency: The agency maintaining the hierarchy for which reports
                must be returned.
            id: The ID of the structure for which reports must be returned.
            version: The version of the structure for which reports must be
                returned.

        Returns:
            The metadata reports about the supplied structure.
        """
        query = super()._reports_q(artefact_type, agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.report, True)

    def get_mapping(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> StructureMap:
        """Get a mapping definition (aka structure map).

        Args:
            agency: The agency maintaining the structure map.
            id: The ID of the structure map to be returned.
            version: The version of the structure map to be returned. The most
                recent version will be returned, unless specified otherwise.

        Returns:
            The requested mapping definition (aka Structure Map).
        """
        query = super()._mapping_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.mapping)

    def get_code_map(
        self, agency: str, id: str, version: str = "+"
    ) -> Union[MultiRepresentationMap, RepresentationMap]:
        """Get a code map (aka representation map).

        Args:
            agency: The agency maintaining the representation map.
            id: The ID of the representation map to be returned.
            version: The version of the representation map to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested mappings in the representation map.
        """
        query = super()._code_map_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.code_map)

    def get_vtl_transformation_scheme(
        self, agency: str, id: str, version: str = "+"
    ) -> Union[MultiRepresentationMap, RepresentationMap]:
        """Get a VTL transformation scheme.

        Args:
            agency: The agency maintaining the transformation scheme.
            id: The ID of the transformation scheme map to be returned.
            version: The version of the transformation scheme to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested transformation scheme.
        """
        query = super()._vtl_ts_q(agency, id, version)
        out = self.__fetch(query)
        return super()._out(out, self.deser.transformation_scheme)


class AsyncRegistryClient(__BaseRegistryClient):
    """A client to be used to retrieve metadata from the FMR.

    With this client, metadata will be retrieved in a asynchronous fashion.
    """

    def __init__(
        self,
        api_endpoint: str,
        format: StructureFormat = StructureFormat.SDMX_JSON_2_0_0,
        pem: Optional[str] = None,
        timeout: float = 10.0,
    ):
        """Instantiate a new client against the target endpoint.

        Args:
            api_endpoint: The endpoint of the targeted service.
            format: The format the service should use to serialize
                the metadata to be returned. Defaults to SDMX-JSON.
            pem: In case the service exposed a certificate created
                by an unknown certificate authority, you can pass
                a pem file for this authority using this parameter.
            timeout: The maximum number of seconds to wait before
                considering that a request timed out. Defaults to
                10 seconds.
        """
        super().__init__(api_endpoint, format)
        self.__service = AsyncRestService(
            self.api_endpoint,
            API_VERSION,
            structure_format=format,
            schema_format=(
                SchemaFormat.SDMX_JSON_2_0_0_STRUCTURE
                if format == StructureFormat.SDMX_JSON_2_0_0
                else SchemaFormat.FUSION_JSON
            ),
            refmeta_format=(
                RefMetaFormat.SDMX_JSON_2_0_0
                if format == StructureFormat.SDMX_JSON_2_0_0
                else RefMetaFormat.FUSION_JSON
            ),
            pem=pem,
            timeout=timeout,
        )

    async def __fetch(
        self,
        query: Union[
            RefMetaByMetadatasetQuery,
            RefMetaByStructureQuery,
            SchemaQuery,
            StructureQuery,
        ],
    ) -> bytes:
        if isinstance(query, StructureQuery):
            out = await self.__service.structure(query)
            return out
        elif isinstance(query, SchemaQuery):
            out = await self.__service.schema(query)
            return out
        else:
            out = await self.__service.reference_metadata(query)
            return out

    async def __get_hierarchies_for_flow(
        self, agency: str, flow: str, version: str
    ) -> Sequence[HierarchyAssociation]:
        query = super()._hierarchies_for_flow_q(agency, flow, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.hier_assoc)

    async def __get_hierarchies_for_pra(
        self, agency: str, pra: str, version: str
    ) -> Sequence[HierarchyAssociation]:
        query = super()._hierarchies_for_pra_q(agency, pra, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.hier_assoc)

    async def get_agencies(self, agency: str) -> Sequence[Agency]:
        """Get the list of **sub-agencies** for the supplied agency.

        Args:
            agency: The agency maintaining the agency scheme from
                which sub-agencies must be returned.

        Returns:
            The requested list of agencies.
        """
        query = super()._agencies_q(agency)
        out = await self.__fetch(query)
        schemes = super()._out(out, self.deser.agencies)
        return schemes[0].items

    async def get_providers(
        self, agency: str, with_flows: bool = False
    ) -> Sequence[DataProvider]:
        """Get the list of **data providers** for the supplied agency.

        Args:
            agency: The agency maintaining the data provider scheme from
                which data providers must be returned.
            with_flows: Whether the data providers should contain the list
                of dataflows for which the data provider provides data.

        Returns:
            The requested list of data providers.
        """
        query = super()._providers_q(agency, with_flows)
        out = await self.__fetch(query)
        schemes = super()._out(out, self.deser.providers)
        return schemes[0].items

    async def get_categories(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> CategoryScheme:
        """Get the category scheme matching the supplied parameters.

        Args:
            agency: The agency maintaining the category scheme.
            id: The ID of the category scheme to be returned.
            version: The version of the category scheme to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested category scheme.
        """
        query = super()._categories_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.categories)

    async def get_categorisation(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Categorisation:
        """Get the categorisation matching the supplied parameters.

        Args:
            agency: The agency maintaining the categorisation.
            id: The ID of the categorisation to be returned.
            version: The version of the categorisation to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested categorisation.
        """
        query = super()._categorisation_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.categorisation)[0]

    async def get_provision_agreement(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> ProvisionAgreement:
        """Get the provision agreement matching the supplied parameters.

        Args:
            agency: The agency maintaining the provision agreement.
            id: The ID of the provision agreement to be returned.
            version: The version of the provision agreement to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested provision agreement.
        """
        query = super()._pa_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.provision_agreement)[0]

    async def get_codes(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Codelist:
        """Get the codes from the codelist matching the supplied parameters.

        Args:
            agency: The agency maintaining the codelist.
            id: The ID of the codelist to be returned.
            version: The version of the codelist to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested codelist.
        """
        try:
            query = super()._codes_cl_q(agency, id, version)
            out = await self.__fetch(query)
            return super()._out(out, self.deser.codes)
        except NotFound:
            query = super()._codes_vl_q(agency, id, version)
            out = await self.__fetch(query)
            return super()._out(out, self.deser.codes)

    async def get_concepts(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> ConceptScheme:
        """Get the concept scheme matching the supplied parameters.

        Args:
            agency: The agency maintaining the concept scheme.
            id: The ID of the concept scheme to be returned.
            version: The version of the concept scheme to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested concept scheme.
        """
        query = super()._concepts_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.concepts)

    async def get_schema(
        self,
        context: Union[
            SchemaContext,
            Literal["dataflow", "datastructure", "provisionagreement"],
        ],
        agency: str,
        id: str,
        version: str,
    ) -> Schema:
        """Get the schema matching the supplied parameters.

        Args:
            context: The context for which the schema should be generated
                (dataflow, datastructure or provisionagreement). This defines
                the constraints to be applied when generating the schema.
            agency: The ID of the agency maintaining the context.
            id: The ID of the context to be considered.
            version: The version of the context to be considered.

        Returns:
            The requested schema.
        """
        c = (
            context
            if isinstance(context, SchemaContext)
            else SchemaContext(context)
        )
        if c == SchemaContext.DATAFLOW:
            ha = await self.__get_hierarchies_for_flow(agency, id, version)
        elif c == SchemaContext.PROVISION_AGREEMENT:
            ha = await self.__get_hierarchies_for_pra(agency, id, version)
        else:
            ha = ()
        query = super()._schema_q(c, agency, id, version)
        r = await self.__fetch(query)
        return super()._out(
            r, self.deser.schema, c.value, agency, id, version, ha
        )

    async def get_dataflow_details(
        self,
        agency: str,
        id: str,
        version: str = "+",
        detail: Union[
            DataflowDetails, Literal["all", "core", "providers", "schema"]
        ] = DataflowDetails.ALL,
    ) -> DataflowInfo:
        """Get detailed information about a dataflow.

        Args:
            agency: The agency maintaining the dataflow.
            id: The ID of the dataflow to be returned.
            version: The version of the dataflow to be returned. The most
                recent version will be returned, unless specified otherwise.
            detail: The amount of detail to be returned. "core" means only
                core information (ID, name, etc.) will be returned.
                "providers" means core information will be returned, as well
                as the list of organizations providing data for the dataflow.
                "schema" means core information will be returned, as well
                as the schema describing the expected/allowed structure of
                data to be reported against the dataflow. "all" combines all
                3 other options.

        Returns:
            The requested information about a dataflow.
        """
        d = DataflowDetails(detail) if isinstance(detail, str) else detail
        sq, dr = super()._df_details(d)
        if sq:
            schema = await self.get_schema(
                "dataflow",
                agency,
                id,
                version,
            )
            cmps = schema.components
        else:
            cmps = None
        query = super()._dataflow_details_q(agency, id, version, dr)
        out = await self.__fetch(query)
        return super()._out(
            out, self.deser.dataflow_info, cmps, agency, id, version
        )

    async def get_dataflows(
        self,
        agency: str = "*",
        id: str = "*",
        version: str = "+",
    ) -> Sequence[Dataflow]:
        """Get the dataflow(s) matching the supplied parameters.

        Args:
            agency: The agency maintaining the dataflow(s).
            id: The ID of the dataflow(s) to be returned.
            version: The version of the dataflow(s) to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested dataflow(s).
        """
        query = super()._dataflows_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.dataflows)

    async def get_hierarchy(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Hierarchy:
        """Get a hierarchical list of codes.

        Args:
            agency: The agency maintaining the hierarchy.
            id: The ID of the hierarchy to be returned.
            version: The version of the hierarchy to be returned. The most
                recent version will be returned, unless specified otherwise.

        Returns:
            The requested hierarchical list of codes.
        """
        query = super()._hierarchy_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.hierarchy)

    async def get_report(
        self,
        provider: str,
        id: str,
        version: str = "+",
    ) -> MetadataReport:
        """Get a reference metadata report.

        Args:
            provider: The organization which provided the report.
            id: The ID of the report to be returned.
            version: The version of the report to be returned. The most
                recent version will be returned, unless specified otherwise.

        Returns:
            The requested metadata report.
        """
        query = super()._report_q(provider, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.report)[0]

    async def get_reports(
        self,
        artefact_type: str,
        agency: str,
        id: str,
        version: str = "+",
    ) -> Sequence[MetadataReport]:
        """Get the reference metadata reports for the supplied structure.

        Args:
            artefact_type: The type of the structure for which reports must
                be returned.
            agency: The agency maintaining the hierarchy for which reports
                must be returned.
            id: The ID of the structure for which reports must be returned.
            version: The version of the structure for which reports must be
                returned.

        Returns:
            The metadata reports about the supplied structure.
        """
        query = super()._reports_q(artefact_type, agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.report, True)

    async def get_mapping(
        self,
        agency: str,
        id: str,
        version: str = "+",
    ) -> StructureMap:
        """Get a mapping definition (aka structure map).

        Args:
            agency: The agency maintaining the structure map.
            id: The ID of the structure map to be returned.
            version: The version of the structure map to be returned. The most
                recent version will be returned, unless specified otherwise.

        Returns:
            The requested mapping definition (aka Structure Map).
        """
        query = super()._mapping_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.mapping)

    async def get_code_map(
        self, agency: str, id: str, version: str = "+"
    ) -> Union[MultiRepresentationMap, RepresentationMap]:
        """Get a code map (aka representation map).

        Args:
            agency: The agency maintaining the representation map.
            id: The ID of the representation map to be returned.
            version: The version of the representation map to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested mappings in the representation map.
        """
        query = super()._code_map_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.code_map)

    async def get_vtl_transformation_scheme(
        self, agency: str, id: str, version: str = "+"
    ) -> Union[MultiRepresentationMap, RepresentationMap]:
        """Get a VTL transformation scheme.

        Args:
            agency: The agency maintaining the transformation scheme.
            id: The ID of the transformation scheme map to be returned.
            version: The version of the transformation scheme to be returned.
                The most recent version will be returned, unless specified
                otherwise.

        Returns:
            The requested transformation scheme.
        """
        query = super()._vtl_ts_q(agency, id, version)
        out = await self.__fetch(query)
        return super()._out(out, self.deser.transformation_scheme)



"""SDMX 2.0 CSV writer module."""

from copy import copy
from typing import Optional, Sequence

import pandas as pd

from pysdmx.io.csv.sdmx20 import SDMX_CSV_ACTION_MAPPER
from pysdmx.io.pd import PandasDataset


def write(
    datasets: Sequence[PandasDataset], output_path: Optional[str] = None
) -> Optional[str]:
    """Write data to SDMX-CSV 2.0 format.

    Args:
        datasets: List of datasets to write.
          Must have the same components.
        output_path: Path to write the data to.
          If None, the data is returned as a string.

    Returns:
        SDMX CSV data as a string, if output_path is None.
    """
    # Link to pandas.to_csv documentation on sphinx:
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_csv.html

    dataframes = []
    for dataset in datasets:
        # Create a copy of the dataset
        df: pd.DataFrame = copy(dataset.data)

        # Add additional attributes to the dataset
        for k, v in dataset.attributes.items():
            df[k] = v

        structure_ref, unique_id = dataset.short_urn.split("=", maxsplit=1)
        if structure_ref in ["DataStructure", "Dataflow"]:
            structure_ref = structure_ref.lower()
        else:
            structure_ref = "dataprovision"

        # Insert two columns at the beginning of the data set
        df.insert(0, "STRUCTURE", structure_ref)
        df.insert(1, "STRUCTURE_ID", unique_id)
        action_value = SDMX_CSV_ACTION_MAPPER[dataset.action]
        df.insert(2, "ACTION", action_value)

        dataframes.append(df)

    all_data = pd.concat(dataframes, ignore_index=True, axis=0)

    # Convert the dataset into a csv file
    return all_data.to_csv(output_path, index=False, header=True)



"""Collection of SDMX-JSON schemas for dataflow queries."""

from datetime import datetime
from typing import List, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation
from pysdmx.io.json.sdmxjson2.messages.org import JsonDataProviderScheme
from pysdmx.model import (
    Agency,
    Components,
    Dataflow,
    DataflowInfo,
    DataProvider,
)


class JsonDataflow(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a dataflow."""

    id: str
    name: str
    agency: str
    structure: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None

    def to_model(self) -> Dataflow:
        """Converts a FusionDataflow to a standard dataflow."""
        return Dataflow(
            id=self.id,
            agency=self.agency,
            name=self.name,
            description=self.description,
            version=self.version,
            structure=self.structure,
        )


class JsonDataflows(Struct, frozen=True):
    """SDMX-JSON payload for the list of dataflows."""

    dataflows: Sequence[JsonDataflow]
    dataProviderSchemes: Sequence[JsonDataProviderScheme] = ()

    def __filter(
        self, df: JsonDataflow, agency: str, id_: str, version: str
    ) -> bool:
        if version != "+" and version != "latest":
            return (
                df.agency == agency and df.id == id_ and df.version == version
            )
        else:
            return df.agency == agency and df.id == id_

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        prvs: List[DataProvider] = []
        for dps in self.dataProviderSchemes:
            prvs.extend(dps.dataProviders)
        df = list(
            filter(
                lambda df: self.__filter(df, agency, id_, version),
                self.dataflows,
            )
        )[0]
        return DataflowInfo(
            id=df.id,
            components=components,
            agency=Agency(df.agency),
            name=df.name,
            description=df.description,
            version=df.version,
            providers=prvs,
            dsd_ref=df.structure,
        )

    def to_simple_model(self) -> Sequence[Dataflow]:
        """Returns the requested dataflows."""
        return [df.to_model() for df in self.dataflows]


class JsonDataflowMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataflow queries (with details)."""

    data: JsonDataflows

    def to_model(
        self, components: Components, agency: str, id_: str, version: str
    ) -> DataflowInfo:
        """Returns the requested dataflow details."""
        return self.data.to_model(components, agency, id_, version)


class JsonDataflowsMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataflow queries."""

    data: JsonDataflows

    def to_model(self) -> Sequence[Dataflow]:
        """Returns the requested dataflows."""
        return self.data.to_simple_model()



from typing import Any, Dict, List, Optional, Tuple

import pandas as pd

from pysdmx.errors import Invalid, NotImplemented
from pysdmx.io.xml.sdmx21.__tokens import (
    AGENCY_ID,
    DATASET,
    DIM_OBS,
    HEADER,
    ID,
    REF,
    STR_ID,
    STR_REF,
    STR_USAGE,
    STRUCTURE,
    URN,
    VERSION,
)
from pysdmx.io.xml.utils import add_list
from pysdmx.util import parse_urn

READING_CHUNKSIZE = 50000


def __process_df(
    test_list: List[Dict[str, Any]],
    df: Optional[pd.DataFrame],
    is_end: bool = False,
) -> Any:
    if not is_end and len(test_list) <= READING_CHUNKSIZE:
        return test_list, df
    if df is not None:
        df = pd.concat([df, pd.DataFrame(test_list)], ignore_index=True)
    else:
        df = pd.DataFrame(test_list)

    del test_list[:]

    return test_list, df


def __get_ids_from_structure(element: Dict[str, Any]) -> Any:
    """Gets the agency_id, id and version of the structure.

    Args:
        element: The data hold in the structure.

    Returns:
        If the element is REF, agency_id, id and version may be returned.
        If the element is URN, agency_id, id and version would be taken from
        split function.
    """
    if REF in element:
        agency_id = element[REF][AGENCY_ID]
        id_ = element[REF][ID]
        version = element[REF][VERSION]
        return agency_id, id_, version
    else:
        urn = parse_urn(element[URN])
        return urn.agency, urn.id, urn.version


def __get_elements_from_structure(structure: Dict[str, Any]) -> Any:
    """Gets elements according to the xml type of file.

    Args:
        structure: It can appear in two ways:
            If structure is 'STRUCTURE', it will get
            the ids related to STRUCTURE.
            If structure is 'STR_USAGE', it will get
            the ids related to STR_USAGE.

    Returns:
        The ids contained in the structure will be returned.

    Raises:
        NotImplemented: For Provision Agreement, as it is not implemented.
    """
    if STRUCTURE in structure:
        structure_type = "DataStructure"
        tuple_ids = __get_ids_from_structure(structure[STRUCTURE])

    elif STR_USAGE in structure:
        structure_type = "Dataflow"
        tuple_ids = __get_ids_from_structure(structure[STR_USAGE])
    else:
        raise NotImplemented(
            "Unsupported", "ProvisionAgrement not implemented"
        )
    return tuple_ids + (structure_type,)


def __extract_structure(structure: Any) -> Any:
    """Extracts elements contained in the structure."""
    structure = add_list(structure)
    str_info = {}
    for str_item in structure:
        (agency_id, id_, version, structure_type) = (
            __get_elements_from_structure(str_item)
        )

        str_id = f"{agency_id}:{id_}({version})"

        str_info[str_item[STR_ID]] = {
            DIM_OBS: str_item[DIM_OBS],
            "unique_id": str_id,
            "structure_type": structure_type,
        }

    return str_info


def get_data_objects(dict_info: Dict[str, Any]) -> Tuple[Any, Any]:
    """Parse dataset.

    Args:
        dict_info: Dict.
          XML dictionary (xmltodict).

    Returns:
        A dictionary of datasets.
    """
    str_info = __extract_structure(dict_info[HEADER][STRUCTURE])
    if DATASET not in dict_info:
        dataset_info = []
        for key in dict_info:
            if DATASET in key:
                dataset_info = add_list(dict_info[key])
    else:
        dataset_info = add_list(dict_info[DATASET])

    for d in dataset_info:
        if d[STR_REF] not in str_info:
            raise Invalid(
                f"Dataset Structure Reference {d[STR_REF]} "
                f"not found in the Header"
            )
    return dataset_info, str_info



"""Collection of SDMX-JSON schemas for categories and category schemes."""

from collections import defaultdict
from datetime import datetime
from typing import Dict, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation
from pysdmx.io.json.sdmxjson2.messages.dataflow import JsonDataflow
from pysdmx.model import (
    Agency,
    Categorisation,
    Category,
    CategoryScheme,
    Dataflow,
    DataflowRef,
)
from pysdmx.util import find_by_urn


class JsonCategorisation(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a categorisation."""

    id: str
    name: str
    agency: str
    source: str
    target: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None

    def to_model(self) -> Categorisation:
        """Converts a JsonCategorisation to a standard categorisation."""
        return Categorisation(
            id=self.id,
            agency=self.agency,
            version=self.version,
            source=self.source,
            target=self.target,
            name=self.name,
            description=self.description,
            is_external_reference=self.isExternalReference,
            valid_from=self.validFrom,
            valid_to=self.validTo,
        )


class JsonCategory(Struct, frozen=True):
    """SDMX-JSON payload for a category."""

    id: str  # type: ignore[misc, unused-ignore]
    name: Optional[str] = None
    description: Optional[str] = None
    categories: Sequence["JsonCategory"] = ()

    def to_model(self) -> Category:
        """Converts a FusionCode to a standard code."""
        return Category(
            id=self.id,
            name=self.name,
            description=self.description,
            categories=[c.to_model() for c in self.categories],
        )


class JsonCategoryScheme(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a category scheme."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    categories: Sequence[JsonCategory] = ()

    def to_model(self) -> CategoryScheme:
        """Converts a JsonCodelist to a standard codelist."""
        return CategoryScheme(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            items=[c.to_model() for c in self.categories],
        )


class JsonCategorySchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of category schemes."""

    categorySchemes: Sequence[JsonCategoryScheme]
    categorisations: Sequence[JsonCategorisation] = ()
    dataflows: Sequence[JsonDataflow] = ()


class JsonCategorySchemeMessage(Struct, frozen=True):
    """SDMX-JSON payload for /categoryscheme queries."""

    data: JsonCategorySchemes

    def __group_flows(self) -> defaultdict[str, list[Dataflow]]:
        out: defaultdict[str, list[Dataflow]] = defaultdict(list)
        for c in self.data.categorisations:
            d = find_by_urn(self.data.dataflows, c.source)
            src = c.target[c.target.find(")") + 2 :]
            out[src].append(d.to_model())
        return out

    def __add_flows(
        self, cat: Category, cni: str, cf: Dict[str, list[Dataflow]]
    ) -> None:
        if cat.categories:
            for c in cat.categories:
                self.__add_flows(c, f"{cni}.{c.id}", cf)
        if cni in cf:
            dfrefs = [
                DataflowRef(
                    (
                        df.agency.id
                        if isinstance(df.agency, Agency)
                        else df.agency
                    ),
                    df.id,
                    df.version,
                    df.name,
                )
                for df in cf[cni]
            ]
            cat.dataflows = dfrefs

    def to_model(self) -> CategoryScheme:
        """Returns the requested codelist."""
        cf = self.__group_flows()
        cs = self.data.categorySchemes[0].to_model()
        for c in cs:
            self.__add_flows(c, c.id, cf)
        return cs


class JsonCategorisations(Struct, frozen=True):
    """SDMX-JSON payload for the list of categorisations."""

    categorisations: Sequence[JsonCategorisation]


class JsonCategorisationMessage(Struct, frozen=True):
    """SDMX-JSON payload for /categorisation queries."""

    data: JsonCategorisations

    def to_model(self) -> Sequence[Categorisation]:
        """Returns the requested categorisations."""
        return [cat.to_model() for cat in self.data.categorisations]



"""SDMX 2.1 XML Generic Data reader module."""

from typing import Any, Dict, Sequence

import pandas as pd

from pysdmx.errors import Invalid
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.sdmx21.__tokens import (
    ATTRIBUTES,
    GENERIC,
    ID,
    OBS,
    OBS_DIM,
    OBS_KEY,
    OBS_VALUE_ID,
    OBS_VALUE_XML_TAG,
    SERIES,
    SERIES_KEY,
    STR_REF,
    VALUE,
)
from pysdmx.io.xml.sdmx21.reader.__data_aux import (
    __process_df,
    get_data_objects,
)
from pysdmx.io.xml.sdmx21.reader.__parse_xml import parse_xml
from pysdmx.io.xml.utils import add_list
from pysdmx.model.dataset import ActionType


def __get_element_to_list(data: Dict[str, Any], mode: Any) -> Dict[str, Any]:
    obs = {}
    data[mode][VALUE] = add_list(data[mode][VALUE])
    for k in data[mode][VALUE]:
        obs[k[ID]] = k[VALUE.lower()]
    return obs


def __reading_generic_series(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Generic Series
    test_list = []
    df = None
    dataset[SERIES] = add_list(dataset[SERIES])
    for series in dataset[SERIES]:
        keys = {}
        # Series Keys
        series[SERIES_KEY][VALUE] = add_list(series[SERIES_KEY][VALUE])
        for v in series[SERIES_KEY][VALUE]:
            keys[v[ID]] = v[VALUE.lower()]
        if ATTRIBUTES in series:
            series[ATTRIBUTES][VALUE] = add_list(series[ATTRIBUTES][VALUE])
            for v in series[ATTRIBUTES][VALUE]:
                keys[v[ID]] = v[VALUE.lower()]
        if OBS in series:
            series[OBS] = add_list(series[OBS])

            for data in series[OBS]:
                obs = {
                    OBS_DIM: data[OBS_DIM][VALUE.lower()],
                    OBS_VALUE_ID: data[OBS_VALUE_XML_TAG][VALUE.lower()],
                }
                if ATTRIBUTES in data:
                    obs = {
                        **obs,
                        **__get_element_to_list(data, mode=ATTRIBUTES),
                    }
                test_list.append({**keys, **obs})
        else:
            test_list.append(keys)
        test_list, df = __process_df(test_list, df)

    test_list, df = __process_df(test_list, df, is_end=True)

    return df


def __reading_generic_all(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Generic All Dimensions
    test_list = []
    if OBS not in dataset:
        return pd.DataFrame()
    df = None
    dataset[OBS] = add_list(dataset[OBS])
    for data in dataset[OBS]:
        obs: Dict[str, Any] = {}
        obs = {
            **obs,
            **__get_element_to_list(data, mode=OBS_KEY),
            OBS_VALUE_ID: data[OBS_VALUE_XML_TAG][VALUE.lower()],
        }
        if ATTRIBUTES in data:
            obs = {**obs, **__get_element_to_list(data, mode=ATTRIBUTES)}
        test_list.append({**obs})
        test_list, df = __process_df(test_list, df)

    test_list, df = __process_df(test_list, df, is_end=True)

    return df


def __get_at_att_gen(dataset: Dict[str, Any]) -> Dict[str, Any]:
    """Gets all the elements if it is Generic data."""
    attached_attributes: Dict[str, Any] = {}
    if ATTRIBUTES not in dataset:
        return attached_attributes
    dataset[ATTRIBUTES][VALUE] = add_list(dataset[ATTRIBUTES][VALUE])
    for k in dataset[ATTRIBUTES][VALUE]:
        attached_attributes[k[ID]] = k[VALUE.lower()]
    return attached_attributes


def __parse_generic_data(
    dataset: Dict[str, Any], structure_info: Dict[str, Any]
) -> PandasDataset:
    attached_attributes = __get_at_att_gen(dataset)

    # Parsing data
    if SERIES in dataset:
        # Generic Series
        df = __reading_generic_series(dataset)
        dim_at_obs = structure_info["dimensionAtObservation"]
        # In case there are observations defined, we need to replace the
        # OBS_DIM column with the dimension at observation
        if OBS_DIM in df.columns:
            df[dim_at_obs] = df[OBS_DIM]
            del df[OBS_DIM]
    else:
        # Generic All Dimensions
        df = __reading_generic_all(dataset)

    action = dataset.get("action", "Information")
    action = ActionType(action)

    urn = f"{structure_info['structure_type']}={structure_info['unique_id']}"

    return PandasDataset(
        structure=urn, attributes=attached_attributes, data=df, action=action
    )


def read(input_str: str, validate: bool = True) -> Sequence[PandasDataset]:
    """Reads an SDMX-ML 2.1 Generic data and returns a Sequence of Datasets.

    Args:
        input_str: SDMX-ML data to read.
        validate: If True, the XML data will be validated against the XSD.
    """
    dict_info = parse_xml(input_str, validate=validate)
    if GENERIC not in dict_info:
        raise Invalid("This SDMX document is not SDMX-ML 2.1 Generic.")
    dataset_info, str_info = get_data_objects(dict_info[GENERIC])

    datasets = []
    for dataset in dataset_info:
        ds = __parse_generic_data(dataset, str_info[dataset[STR_REF]])
        datasets.append(ds)
    return datasets



"""Connector to SDMX-REST services."""

from typing import NoReturn, Optional, Union

import httpx

from pysdmx import errors
from pysdmx.api.qb.availability import AvailabilityFormat, AvailabilityQuery
from pysdmx.api.qb.data import DataFormat, DataQuery
from pysdmx.api.qb.refmeta import (
    RefMetaByMetadataflowQuery,
    RefMetaByMetadatasetQuery,
    RefMetaByStructureQuery,
    RefMetaFormat,
)
from pysdmx.api.qb.registration import (
    RegistrationByContextQuery,
    RegistrationByIdQuery,
    RegistrationByProviderQuery,
    RegistryFormat,
)
from pysdmx.api.qb.schema import SchemaFormat, SchemaQuery
from pysdmx.api.qb.structure import StructureFormat, StructureQuery
from pysdmx.api.qb.util import ApiVersion


class _CoreRestService:
    """Abstract connector."""

    def __init__(
        self,
        api_endpoint: str,
        api_version: ApiVersion,
        data_format: DataFormat,
        structure_format: StructureFormat,
        schema_format: SchemaFormat,
        refmeta_format: RefMetaFormat,
        avail_format: AvailabilityFormat,
        registry_format: RegistryFormat,
        pem: Optional[str] = None,
        timeout: Optional[float] = 5.0,
    ):
        """Instantiate a connector to a SDMX-REST service."""
        self._api_endpoint = (
            api_endpoint[0:-1] if api_endpoint.endswith("/") else api_endpoint
        )

        self._api_version = api_version
        self._data_format = data_format
        self._structure_format = structure_format
        self._schema_format = schema_format
        self._refmeta_format = refmeta_format
        self._avail_format = avail_format
        self._registry_format = registry_format
        self._ssl_context = (
            httpx.create_ssl_context(
                verify=pem,
            )
            if pem
            else httpx.create_ssl_context()
        )
        self._headers = {
            "Accept-Encoding": "gzip, deflate",
        }
        self._timeout = timeout

    def _map_error(
        self, e: Union[httpx.RequestError, httpx.HTTPStatusError]
    ) -> NoReturn:
        q = e.request.url
        if isinstance(e, httpx.HTTPStatusError):
            s = e.response.status_code
            t = e.response.text
            if s == 404:
                msg = (
                    "The requested resource(s) could not be found in the "
                    f"targeted service. The query was `{q}`"
                )
                raise errors.NotFound("Not found", msg) from e
            elif s < 500:
                msg = (
                    f"The query returned a {s} error code. The query "
                    f"was `{q}`. The error message was: `{t}`."
                )
                raise errors.Invalid(f"Client error {s}", msg) from e
            else:
                msg = (
                    f"The service returned a {s} error code. The query "
                    f"was `{q}`. The error message was: `{t}`."
                )
                raise errors.InternalError(f"Service error {s}", msg) from e
        else:
            msg = (
                f"There was an issue connecting to the targeted service. "
                f"The query was `{q}`. The error message was: `{e}`."
            )
            raise errors.Unavailable("Connection error", msg) from e


class RestService(_CoreRestService):
    """Synchronous connector to SDMX-REST services.

    Attributes:
        api_endpoint: The entry point (URL) of the SDMX-REST service.
        api_version: The most recent version of the SDMX-REST specification
            supported by the service.
        data_format: The default format for data queries.
        structure_format: The default format for structure queries.
        schema_format: The default format for schema queries.
        refmeta_format: The default format for reference metadata queries.
        avail_format: The default format for availability queries.
        registry_format: The default format for registration queries.
        pem: In case the service uses SSL/TLS with self-signed certificate,
            this attribute should be used to pass the pem file with the
            list of trusted certicate authorities.
        timeout: The maximum number of seconds to wait before considering
            that a request timed out. Defaults to 5 seconds.
    """

    def __init__(
        self,
        api_endpoint: str,
        api_version: ApiVersion,
        data_format: DataFormat = DataFormat.SDMX_JSON_2_0_0,
        structure_format: StructureFormat = StructureFormat.SDMX_JSON_2_0_0,
        schema_format: SchemaFormat = SchemaFormat.SDMX_JSON_2_0_0_STRUCTURE,
        refmeta_format: RefMetaFormat = RefMetaFormat.SDMX_JSON_2_0_0,
        avail_format: AvailabilityFormat = AvailabilityFormat.SDMX_JSON_2_0_0,
        registry_format: RegistryFormat = RegistryFormat.FUSION_JSON,
        pem: Optional[str] = None,
        timeout: Optional[float] = 5.0,
    ):
        """Instantiate a connector to a SDMX-REST service."""
        super().__init__(
            api_endpoint,
            api_version,
            data_format,
            structure_format,
            schema_format,
            refmeta_format,
            avail_format,
            registry_format,
            pem,
            timeout,
        )

    def data(self, query: DataQuery) -> bytes:
        """Execute a data query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._data_format.value
        return self.__fetch(q, f)

    def structure(self, query: StructureQuery) -> bytes:
        """Execute a structure query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._structure_format.value
        return self.__fetch(q, f)

    def schema(self, query: SchemaQuery) -> bytes:
        """Execute a schema query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._schema_format.value
        return self.__fetch(q, f)

    def availability(self, query: AvailabilityQuery) -> bytes:
        """Execute an availability query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._avail_format.value
        return self.__fetch(q, f)

    def reference_metadata(
        self,
        query: Union[
            RefMetaByMetadataflowQuery,
            RefMetaByMetadatasetQuery,
            RefMetaByStructureQuery,
        ],
    ) -> bytes:
        """Execute a reference metadata query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._refmeta_format.value
        return self.__fetch(q, f)

    def registration(
        self,
        query: Union[
            RegistrationByContextQuery,
            RegistrationByIdQuery,
            RegistrationByProviderQuery,
        ],
    ) -> bytes:
        """Execute a registration query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._registry_format.value
        return self.__fetch(q, f)

    def __fetch(self, query: str, format: str) -> bytes:
        with httpx.Client(verify=self._ssl_context) as client:
            try:
                url = f"{self._api_endpoint}{query}"
                h = self._headers.copy()
                h["Accept"] = format
                r = client.get(url, headers=h, timeout=self._timeout)
                r.raise_for_status()
                return r.content
            except (httpx.RequestError, httpx.HTTPStatusError) as e:
                self._map_error(e)


class AsyncRestService(_CoreRestService):
    """Asynchronous connector to SDMX-REST services.

    Attributes:
        api_endpoint: The entry point (URL) of the SDMX-REST service.
        api_version: The most recent version of the SDMX-REST specification
            supported by the service.
        data_format: The default format for data queries.
        structure_format: The default format for structure queries.
        schema_format: The default format for schema queries.
        refmeta_format: The default format for reference metadata queries.
        avail_format: The default format for availability queries.
        registry_format: The default format for registration queries.
        pem: In case the service uses SSL/TLS with self-signed certificate,
            this attribute should be used to pass the pem file with the
            list of trusted certicate authorities.
        timeout: The maximum number of seconds to wait before considering
            that a request timed out. Defaults to 5 seconds.
    """

    def __init__(
        self,
        api_endpoint: str,
        api_version: ApiVersion,
        data_format: DataFormat = DataFormat.SDMX_JSON_2_0_0,
        structure_format: StructureFormat = StructureFormat.SDMX_JSON_2_0_0,
        schema_format: SchemaFormat = SchemaFormat.SDMX_JSON_2_0_0_STRUCTURE,
        refmeta_format: RefMetaFormat = RefMetaFormat.SDMX_JSON_2_0_0,
        avail_format: AvailabilityFormat = AvailabilityFormat.SDMX_JSON_2_0_0,
        registry_format: RegistryFormat = RegistryFormat.FUSION_JSON,
        pem: Optional[str] = None,
        timeout: Optional[float] = 5.0,
    ):
        """Instantiate a connector to a SDMX-REST service."""
        super().__init__(
            api_endpoint,
            api_version,
            data_format,
            structure_format,
            schema_format,
            refmeta_format,
            avail_format,
            registry_format,
            pem,
            timeout,
        )

    async def data(self, query: DataQuery) -> bytes:
        """Execute a data query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._data_format.value
        out = await self.__fetch(q, f)
        return out

    async def structure(self, query: StructureQuery) -> bytes:
        """Execute a structure query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._structure_format.value
        out = await self.__fetch(q, f)
        return out

    async def schema(self, query: SchemaQuery) -> bytes:
        """Execute a schema query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._schema_format.value
        out = await self.__fetch(q, f)
        return out

    async def availability(self, query: AvailabilityQuery) -> bytes:
        """Execute an availability query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._avail_format.value
        out = await self.__fetch(q, f)
        return out

    async def reference_metadata(
        self,
        query: Union[
            RefMetaByMetadataflowQuery,
            RefMetaByMetadatasetQuery,
            RefMetaByStructureQuery,
        ],
    ) -> bytes:
        """Execute a reference metadata query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._refmeta_format.value
        out = await self.__fetch(q, f)
        return out

    async def registration(
        self,
        query: Union[
            RegistrationByContextQuery,
            RegistrationByIdQuery,
            RegistrationByProviderQuery,
        ],
    ) -> bytes:
        """Execute a registration query against the service."""
        q = query.get_url(self._api_version, True)
        f = self._registry_format.value
        out = await self.__fetch(q, f)
        return out

    async def __fetch(self, query: str, format: str) -> bytes:
        async with httpx.AsyncClient(verify=self._ssl_context) as client:
            try:
                url = f"{self._api_endpoint}{query}"
                h = self._headers.copy()
                h["Accept"] = format
                r = await client.get(url, headers=h, timeout=self._timeout)
                r.raise_for_status()
                return r.content
            except (httpx.RequestError, httpx.HTTPStatusError) as e:
                self._map_error(e)



"""Utility classes for pysdmx REST query builders."""

import re
from abc import abstractmethod
from enum import IntEnum
from typing import Any, Sequence, Union

import msgspec

from pysdmx.errors import Invalid


class ApiVersion(IntEnum):
    """The version of the SDMX-REST API."""

    V1_0_0 = 0
    V1_0_1 = 1
    V1_0_2 = 2
    V1_1_0 = 3
    V1_2_0 = 4
    V1_3_0 = 5
    V1_4_0 = 6
    V1_5_0 = 7
    V2_0_0 = 8
    V2_1_0 = 9
    V2_2_0 = 10


MULT_SEP = re.compile(r"\+")
REST_ALL = "*"
REST_LATEST = "~"


def check_multiple_items(
    value: Union[str, Sequence[str]], version: ApiVersion
) -> None:
    """Whether multiple items are supported in the supplied API version."""
    if not isinstance(value, str) and version < ApiVersion.V1_3_0:
        raise Invalid(
            "Validation Error",
            f"Multiple items not allowed in SDMX-REST {version.value}.",
        )


def check_multiple_data_context(
    field: str, value: Union[str, Sequence[str]], version: ApiVersion
) -> None:
    """Whether multiple items are supported in the supplied API version."""
    if not isinstance(value, str) and version < ApiVersion.V2_0_0:
        raise Invalid(
            "Validation Error",
            (
                f"More than one {field} is not allowed in data context "
                f"for SDMX-REST {version.value}."
            ),
        )


class CoreQuery(msgspec.Struct, frozen=True, omit_defaults=True):
    def get_url(self, version: ApiVersion, omit_defaults: bool = False) -> str:
        """The URL for the query in the selected SDMX-REST API version."""
        self._validate_query(version)
        if omit_defaults:
            return self._create_short_query(version)
        else:
            return self._create_full_query(version)

    def validate(self) -> None:
        """Validate the query."""
        try:
            self._get_decoder().decode(_encoder.encode(self))
        except msgspec.DecodeError as err:
            raise Invalid("Invalid Schema Query", str(err)) from err

    def _to_kw(self, val: str, ver: ApiVersion) -> str:
        if val == "*" and ver < ApiVersion.V2_0_0:
            val = "all"
        elif val == "~" and ver < ApiVersion.V2_0_0:
            val = "latest"
        return val

    def _to_kws(self, vals: Union[str, Sequence[str]], ver: ApiVersion) -> str:
        vals = [vals] if isinstance(vals, str) else vals
        mapped = [self._to_kw(v, ver) for v in vals]
        return ",".join(mapped)

    def _join_mult(self, vals: Union[str, Sequence[str]]) -> str:
        return vals if isinstance(vals, str) else ",".join(vals)

    def _append_qs_param(
        self, qs: str, value: Any, field: str, disp_value: Any = None
    ) -> str:
        if value or (not isinstance(value, bool) and value == 0):
            if qs:
                qs += "&"
            qs += f"{field}={disp_value if disp_value else value}"
        return qs

    @abstractmethod
    def _get_decoder(self) -> msgspec.json.Decoder:  # type: ignore[type-arg]
        """Returns the decoder to be used for validation."""

    @abstractmethod
    def _validate_query(self, version: ApiVersion) -> None:
        """Any additional validation steps to be performed by subclasses."""

    @abstractmethod
    def _create_full_query(self, ver: ApiVersion) -> str:
        """Creates a URL, with default values."""

    @abstractmethod
    def _create_short_query(self, ver: ApiVersion) -> str:
        """Creates a URL, omitting default values when possible."""


_encoder = msgspec.json.Encoder()

__all__ = ["ApiVersion", "check_multiple_items", "check_multiple_data_context"]



"""Collection of SDMX-JSON schemas for content constraints."""

from datetime import datetime
from typing import Dict, Literal, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import JsonAnnotation


class JsonValue(Struct, frozen=True):
    """SDMX-JSON payload for an allowed value."""

    value: str


class JsonKeyValue(Struct, frozen=True):
    """SDMX-JSON payload for the list of allowed values per component."""

    id: str
    values: Sequence[JsonValue]

    def to_model(self) -> Sequence[str]:
        """Returns the requested list of values."""
        return [v.value for v in self.values]


class JsonCubeRegion(Struct, frozen=True):
    """SDMX-JSON payload for a cube region."""

    keyValues: Sequence[JsonKeyValue]

    def to_map(self) -> Dict[str, Sequence[str]]:
        """Gets the list of allowed values for a component."""
        return {kv.id: kv.to_model() for kv in self.keyValues}


class JsonConstraintAttachment(Struct, frozen=True):
    """SDMX-JSON payload for a constraint attachment."""

    dataProvider: Optional[str]
    simpleDataSources: Optional[Sequence[str]] = None
    dataStructures: Optional[Sequence[str]] = None
    dataflows: Optional[Sequence[str]] = None
    provisionAgreements: Optional[Sequence[str]] = None
    queryableDataSources: Optional[Sequence[str]] = None


class JsonDataConstraint(Struct, frozen=True, rename={"agency": "agencyID"}):
    """SDMX-JSON payload for a content constraint."""

    id: str
    name: str
    agency: str
    description: Optional[str] = None
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    isPartial: bool = False
    role: Optional[Literal["Allowed", "Actual"]] = None
    constraintAttachment: Optional[JsonConstraintAttachment] = None
    cubeRegions: Optional[Sequence[JsonCubeRegion]] = None



"""Model for SDMX Codes and Codelists.

Two different types of collections are provided:

- Codelist: A flat structure, i.e. a structure where all codes are at the same
  level. In SDMX terms, this structure can be fed from information stored
  as a *codelist* or a *value list*.
- Hierarchy: A structure where codes can be organized in a hierarchical
  fashion, like a list of country groups and their respective countries.

This is an example of what is meant with pysdmx being opinionated.
While SDMX allows two different types of hierarchies (simple hierarchies
expressed as codelists, and more complex hierarchies expressed as SDMX
hierarchies), pysdmx expects codelists and value lists to be flat (which
seems to be the vast majority of the cases), thereby leaving the
representation of hierarchical relationships to hierarchies only.
"""

from datetime import datetime
from typing import Iterator, Literal, Optional, Sequence

from msgspec import Struct

from pysdmx.model.__base import Item, ItemScheme, MaintainableArtefact


class Code(Item, frozen=True, omit_defaults=True):
    """A code, such as a country code in the list of ISO 3166 codes.

    Codes may have business validity information.

    Attributes:
        id: The identifier for the code (e.g. UY).
        name: The code's name (e.g. Uruguay).
        description: Additional descriptive information about the code.
        valid_from: Start of the code's validity period.
        valid_to: End of the code's validity period.
    """

    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None


class Codelist(ItemScheme, frozen=True, omit_defaults=True, tag=True):
    """An immutable collection of codes, such as the ISO 3166 country codes.

    A codelist is **maintained by its agency**, typically, an organisation
    like the BIS, the ECB, the IMF, etc.

    A codelist has an **identifier** and a **name**. It may also have a
    **description** and **business validity information**.

    A codelist is **iterable**, i.e. it can be used as is in a for loop.

    A codelist represents a flat list of codes, without any hierarchical
    relationship between them. Both SDMX codelists and SDMX value lists
    are represented as a Codelist class.

    Attributes:
        id: The identifier for the codelist (e.g. CL_FREQ).
        name: The codelist name (e.g. "Frequency codelist").
        agency: The maintainer of the codelist (e.g. SDMX).
        description: Additional descriptive information about the codelist
            (e.g. "This codelist provides a set of values indicating the
            frequency of the data").
        version: The codelist version (e.g. 2.0.42)
    """

    items: Sequence[Code] = ()
    sdmx_type: Literal["codelist", "valuelist"] = "codelist"

    @property
    def codes(self) -> Sequence[Code]:
        """Extract the items in the Codelist."""
        return self.items

    def __iter__(self) -> Iterator[Code]:
        """Return an iterator over the list of codes."""
        yield from self.codes

    def __len__(self) -> int:
        """Return the number of codes in the codelist."""
        return len(self.codes)

    def __getitem__(self, id_: str) -> Optional[Code]:
        """Return the code identified by the supplied ID."""
        out = list(filter(lambda code: code.id == id_, self.codes))
        if len(out) == 0:
            return None
        else:
            return out[0]

    def __contains__(self, id_: str) -> bool:
        """Whether a code with the supplied ID is present in the codelist."""
        return bool(self.__getitem__(id_))


class HierarchicalCode(Struct, frozen=True, omit_defaults=True):
    """A code, as used in a hierarchy.

    Hierachical codes may contain other codes.

    As codes, hierarchial codes may have business validity information.
    In addition, they may also have business validity associated with
    their relationship in the hierarchy.

    For example, let's imagine a hierarchy representing country groups.
    A country might have a valid_from property, with the date that
    country declared its independence. In addition, it may also
    have a rel_valid_from property, with the date the country joined
    that particular group (e.g. European Union).

    Attributes:
        id: The identifier for the code (e.g. UY).
        name: The code's name (e.g. Uruguay).
        description: Additional descriptive information about the code.
        valid_from: Start of the code's validity period.
        valid_to: End of the code's validity period.
        rel_valid_from: Start of the hierarchical relationship validity.
        rel_valid_to: End of the hierarchical relationship validity.
        codes: The child codes.
    """

    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None
    rel_valid_from: Optional[datetime] = None
    rel_valid_to: Optional[datetime] = None
    codes: Sequence["HierarchicalCode"] = ()

    def __iter__(self) -> Iterator["HierarchicalCode"]:
        """Return an iterator over the list of codes."""
        yield from self.codes

    def __str__(self) -> str:
        """Returns a human-friendly description."""
        out = self.id
        if self.name:
            out = f"{out} ({self.name})"
        return out


class Hierarchy(
    MaintainableArtefact, frozen=True, omit_defaults=True, tag=True
):
    """An immutable collection of codes, organized hierarchically.

    A hierarchy is **maintained by its agency**, typically, an organisation
    like the BIS, the ECB, the IMF, etc.

    A hierarchy has an **identifier** and a **name**. It may also have a
    **description** and **business validity information**.

    A hierarchy is **iterable**, i.e. it can be used as is in a for loop.

    Attributes:
        id: The identifier for the hierarchy (e.g. AREA).
        name: The hierarchy name (e.g. "Country groups and their composition").
        agency: The maintainer of the hierarchy (e.g. SDMX).
        description: Additional descriptive information about the hierarchy
            (e.g. "This hierarchy provides a set of country groups and their
            respective composition").
        version: The hierarchy version (e.g. 2.0.42)
        codes: The list of codes in the hierarchy.
        operator: The URN of the operator to be applied to the items of an
            hierarchy. This is mainly used for data validation or data
            compilation purposes. For example, Let's assume a hierarchy with
            a top level code (A), with 2 child codes (B and C). And let's
            assume that the operator property references a VTL operator
            representing a sum. This can then be used for validation purposes,
            to check that A = B + C.
    """

    codes: Sequence[HierarchicalCode] = ()
    operator: Optional[str] = None

    def __iter__(self) -> Iterator[HierarchicalCode]:
        """Return an iterator over the list of codes."""
        yield from self.codes

    def __len__(self) -> int:
        """Return the number of codes in the hierarchy."""
        return self.__get_count(self.codes)

    def __getitem__(self, id_: str) -> Optional[HierarchicalCode]:
        """Return the code identified by the supplied ID."""
        return self.__extract_code(self.codes, id_)

    def __contains__(self, id_: str) -> bool:
        """Whether a code with the supplied ID is present in the hierarchy."""
        return bool(self.__getitem__(id_))

    def __get_count(self, codes: Sequence[HierarchicalCode]) -> int:
        """Return the number of codes at all levels."""
        count = len(codes)
        for code in codes:
            if code.codes:
                count += self.__get_count(code.codes)
        return count

    def __extract_code(
        self, codes: Sequence[HierarchicalCode], id_: str
    ) -> Optional[HierarchicalCode]:
        if "." in id_:
            ids = id_.split(".")
            out = list(filter(lambda c: c.id == ids[0], codes))
            if out:
                pkey = ".".join(ids[1:])
                return self.__extract_code(out[0].codes, pkey)
        else:
            out = list(filter(lambda c: c.id == id_, codes))
            if out:
                return out[0]
        return None

    def __by_id(
        self,
        id: str,
        codes: Sequence[HierarchicalCode],
        out: Optional[Sequence[HierarchicalCode]] = None,
    ) -> Sequence[HierarchicalCode]:
        if out is None:
            out = []
        for i in codes:
            if i.id == id and i not in out:
                out.append(i)  # type: ignore[attr-defined]
            if i.codes:
                self.__by_id(id, i.codes, out)
        return out

    def by_id(self, id: str) -> Sequence[HierarchicalCode]:
        """Get a code without knowing its parent IDs.

        Codes in a hierarchy can be retrieved using their full ID,
        i.e. the code ID, as well as the IDs of their parents in the
        hierarchy, separated by dots (e.g. 1.11.111).

        This function can be used when you just know the code ID,
        and not the ID of its parents.

        Args:
            id: The ID of the code to be returned.

        Returns:
            A set with the matching codes. If there is no matching
            code, the set will be empty. If a code is attached to
            multiple parents, it will be returned only once. As
            hierarchies can reference codes from multiple codelists,
            we could have different codes with the same ID in the
            returned set.
        """
        return self.__by_id(id, self.codes)

    def __get_codes(
        self, codes: Sequence[HierarchicalCode]
    ) -> Sequence[HierarchicalCode]:
        out = []
        for code in codes:
            out.append(code)
            if code.codes:
                out.extend(self.__get_codes(code.codes))
        return out

    def all_codes(self) -> Sequence[HierarchicalCode]:
        """Get all the codes in the hierarchy as a flat list.

        This is useful for validation purposes. The sequence behaves
        as a set, i.e. even if a code is attached to multiple nodes,
        it will be available only once in the returned sequence.

        Returns:
            A flat list of the codes present in the hierarchy.
        """
        out = []
        # We need to do this below because a hierarchical code is not
        # (yet?) hashable.
        for c in self.__get_codes(self.codes):
            if c not in out:
                out.append(c)
        return out


class HierarchyAssociation(
    MaintainableArtefact, frozen=True, omit_defaults=True
):
    """Links a hierarchy to a component withing the context of a dataflow."""

    hierarchy: Optional[Hierarchy] = None
    component_ref: str = ""
    context_ref: str = ""
    operator: Optional[str] = None



"""Utility functions for XML parsing and serialization."""

from typing import Any, List


def add_list(element: Any) -> List[Any]:
    """Make sure an element is a list and convert it if it is not.

    The xmltodict library always returns a list if there are multiple elements
    but a dict if there is only one element. This function makes sure that the
    element is always a list.

    Args:
        element: The element to be converted

    Returns:
        A list with the element
    """
    if not isinstance(element, list):
        element = [element]
    return element



"""Parsers for reading metadata."""

from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.errors import Invalid
from pysdmx.io.xml.sdmx21.__tokens import (
    AGENCIES,
    AGENCY,
    AGENCY_ID,
    ANNOTATION,
    ANNOTATION_TEXT,
    ANNOTATION_TITLE,
    ANNOTATION_TYPE,
    ANNOTATION_URL,
    ANNOTATIONS,
    AS_STATUS,
    ATT,
    ATT_LIST,
    ATT_LVL,
    ATT_REL,
    CL,
    CLASS,
    CLS,
    CODE,
    CODES_LOW,
    COMPS,
    CON,
    CON_ID,
    CON_LOW,
    CON_ROLE,
    CONCEPTS,
    CONTACT,
    CORE_REP,
    CS,
    DEPARTMENT,
    DESC,
    DFW,
    DFWS,
    DIM,
    DIM_LIST,
    DSD,
    DSD_COMPS,
    DSDS,
    DTYPE,
    EMAIL,
    EMAILS,
    ENUM,
    ENUM_FORMAT,
    FACETS,
    FAX,
    FAXES,
    GROUP,
    GROUP_DIM,
    ID,
    IS_EXTERNAL_REF,
    IS_EXTERNAL_REF_LOW,
    IS_FINAL,
    IS_FINAL_LOW,
    IS_PARTIAL,
    IS_PARTIAL_LOW,
    LOCAL_CODES_LOW,
    LOCAL_DTYPE,
    LOCAL_FACETS_LOW,
    LOCAL_REP,
    MANDATORY,
    ME_LIST,
    NAME,
    ORGS,
    PAR_ID,
    PAR_VER,
    PRIM_MEASURE,
    REF,
    REQUIRED,
    ROLE,
    RULE,
    RULE_SCHEME,
    RULESETS,
    SER_URL,
    SER_URL_LOW,
    STR_URL,
    STR_URL_LOW,
    STRUCTURE,
    STRUCTURES,
    TELEPHONE,
    TELEPHONES,
    TEXT,
    TEXT_FORMAT,
    TEXT_TYPE,
    TIME_DIM,
    TITLE,
    TRANS_SCHEME,
    TRANSFORMATION,
    TRANSFORMATIONS,
    TYPE,
    UDO,
    UDO_SCHEME,
    UDOS,
    URI,
    URIS,
    URL,
    URN,
    VALID_FROM,
    VALID_FROM_LOW,
    VALID_TO,
    VALID_TO_LOW,
    VERSION,
)
from pysdmx.io.xml.sdmx21.reader.__parse_xml import parse_xml
from pysdmx.io.xml.utils import add_list
from pysdmx.model import (
    AgencyScheme,
    Code,
    Codelist,
    Concept,
    ConceptScheme,
    DataType,
    Facets,
)
from pysdmx.model.__base import (
    Agency,
    Annotation,
    Contact,
    Item,
    ItemReference,
    ItemScheme,
    Reference,
)
from pysdmx.model.dataflow import (
    Component,
    Components,
    Dataflow,
    DataStructureDefinition,
    Role,
)
from pysdmx.model.vtl import (
    Ruleset,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
)
from pysdmx.util import find_by_urn, parse_urn

STRUCTURES_MAPPING = {
    CL: Codelist,
    AGENCIES: AgencyScheme,
    CS: ConceptScheme,
    DFWS: Dataflow,
    DSDS: DataStructureDefinition,
    RULE_SCHEME: RulesetScheme,
    UDO_SCHEME: UserDefinedOperatorScheme,
    TRANS_SCHEME: TransformationScheme,
}
ITEMS_CLASSES = {
    AGENCY: Agency,
    CODE: Code,
    CON: Concept,
    RULE: Ruleset,
    UDO: UserDefinedOperator,
    TRANSFORMATION: Transformation,
}

COMP_TYPES = [DIM, ATT, PRIM_MEASURE, GROUP_DIM]

ROLE_MAPPING = {
    DIM: Role.DIMENSION,
    ATT: Role.ATTRIBUTE,
    PRIM_MEASURE: Role.MEASURE,
}

FACETS_MAPPING = {
    "minLength": "min_length",
    "maxLength": "max_length",
    "minValue": "min_value",
    "maxValue": "max_value",
    "startValue": "start_value",
    "endValue": "end_value",
    "interval": "interval",
    "timeInterval": "time_interval",
    "decimals": "decimals",
    "pattern": "pattern",
    "startTime": "start_time",
    "endTime": "end_time",
    "isSequence": "is_sequence",
}


def _extract_text(element: Any) -> str:
    """Extracts the text from the element.

    Args:
        element: The element to extract the text from

    Returns:
        The text extracted from the element
    """
    if isinstance(element, list):
        aux = {}
        for language_element in element:
            if "lang" in language_element and language_element["lang"] == "en":
                aux = language_element
        if not aux:
            aux = element[0]
        element = aux
    if isinstance(element, dict) and "#text" in element:
        element = element["#text"]
    return element


class StructureParser(Struct):
    """StructureParser class for SDMX-ML 2.1."""

    agencies: Dict[str, Any] = {}
    codelists: Dict[str, Any] = {}
    concepts: Dict[str, Any] = {}
    datastructures: Dict[str, Any] = {}
    dataflows: Dict[str, Any] = {}
    rulesets: Dict[str, Any] = {}
    udos: Dict[str, Any] = {}
    transformations: Dict[str, Any] = {}

    def __format_contact(self, json_contact: Dict[str, Any]) -> Contact:
        """Creates a Contact object from a json_contact.

        Args:
            json_contact: The element to create the Contact object from

        Returns:
            Contact object created from the json_contact
        """
        self.__format_name_description(json_contact)

        xml_node_to_attribute = {
            NAME: NAME.lower(),
            DEPARTMENT: DEPARTMENT.lower(),
            ROLE: ROLE.lower(),
            URI: URIS,
            EMAIL: EMAILS,
            TELEPHONE: TELEPHONES,
            FAX: FAXES,
        }

        for k, v in xml_node_to_attribute.items():
            if k in json_contact:
                if k in [DEPARTMENT, ROLE]:
                    json_contact[v] = _extract_text(json_contact.pop(k))
                    continue
                field_info = add_list(json_contact.pop(k))
                for i, element in enumerate(field_info):
                    field_info[i] = _extract_text(element)
                json_contact[v] = field_info

        return Contact(**json_contact)

    @staticmethod
    def __format_annotations(item_elem: Any) -> Dict[str, Any]:
        """Formats the annotations in this element.

        Args:
            item_elem: The element to be formatted

        Returns:
            annotations formatted
        """
        if ANNOTATIONS not in item_elem:
            return item_elem
        annotations = []

        ann = item_elem[ANNOTATIONS]
        ann[ANNOTATION] = add_list(ann[ANNOTATION])
        for e in ann[ANNOTATION]:
            if ANNOTATION_TITLE in e:
                e[TITLE] = e.pop(ANNOTATION_TITLE)
            if ANNOTATION_TYPE in e:
                e[TYPE] = e.pop(ANNOTATION_TYPE)
            if ANNOTATION_TEXT in e:
                e[TEXT] = _extract_text(e[ANNOTATION_TEXT])
                del e[ANNOTATION_TEXT]
            if ANNOTATION_URL in e:
                e[URL] = e.pop(ANNOTATION_URL)

            annotations.append(Annotation(**e))

        item_elem[ANNOTATIONS.lower()] = annotations
        del item_elem[ANNOTATIONS]

        return item_elem

    @staticmethod
    def __format_name_description(element: Any) -> Dict[str, Any]:
        node = [NAME, DESC]
        for field in node:
            if field in element:
                element[field.lower()] = _extract_text(element[field])
                del element[field]
        return element

    @staticmethod
    def __format_facets(
        json_fac: Dict[str, Any], json_obj: Dict[str, Any]
    ) -> None:
        """Formats the facets from the JSON information.

        Args:
            json_fac: The element with the facets to be formatted
            json_obj: The element to store the formatted facets
        """
        if json_fac is None:
            return
        for key, _value in json_fac.items():
            if key == TEXT_TYPE and json_fac[TEXT_TYPE] in list(DataType):
                json_obj["dtype"] = DataType(json_fac[TEXT_TYPE])

            if key in FACETS_MAPPING:
                facet_kwargs = {
                    FACETS_MAPPING[k]: v
                    for k, v in json_fac.items()
                    if k in FACETS_MAPPING
                }
                json_obj[FACETS.lower()] = Facets(**facet_kwargs)

    @staticmethod
    def __format_validity(element: Dict[str, Any]) -> Dict[str, Any]:
        if VALID_FROM in element:
            element[VALID_FROM_LOW] = datetime.fromisoformat(
                element.pop(VALID_FROM)
            )
        if VALID_TO in element:
            element[VALID_TO_LOW] = datetime.fromisoformat(
                element.pop(VALID_TO)
            )
        return element

    @staticmethod
    def __format_urls(json_elem: Dict[str, Any]) -> Dict[str, Any]:
        """Formats the STR_URL and SER_URL keys in the element.

        Args:
            json_elem: The element to be formatted

        Returns:
            The json_elem with STR_URL and SER_URL keys formatted.
        """
        if STR_URL in json_elem:
            json_elem[STR_URL_LOW] = json_elem.pop(STR_URL)
        if SER_URL in json_elem:
            json_elem[SER_URL_LOW] = json_elem.pop(SER_URL)
        return json_elem

    def __format_agency(self, element: Dict[str, Any]) -> Dict[str, Any]:
        """Formats the AGENCY_ID key in the element to the maintainer.

        Args:
            element: The element with the Agency ID to be formatted

        Returns:
            element with the Agency ID formatted
        """
        element[AGENCY.lower()] = self.agencies.get(
            element[AGENCY_ID], element[AGENCY_ID]
        )
        del element[AGENCY_ID]
        return element

    def __format_orgs(self, json_orgs: Dict[str, Any]) -> Dict[str, Any]:
        orgs: Dict[str, Any] = {}
        json_list = add_list(json_orgs)
        for e in json_list:
            ag_sch = self.__format_scheme(
                e,
                AGENCIES,
                AGENCY,
            )
            orgs = {**orgs, **ag_sch}
        return orgs

    def __format_representation(
        self, json_rep: Dict[str, Any], json_obj: Dict[str, Any]
    ) -> None:
        """Formats the representation in the JSON Representation."""
        if TEXT_FORMAT in json_rep:
            self.__format_facets(json_rep[TEXT_FORMAT], json_obj)

        if ENUM in json_rep and len(self.codelists) > 0:
            ref = json_rep[ENUM].get(REF, json_rep[ENUM])

            if "URN" in ref:
                codelist = find_by_urn(
                    list(self.codelists.values()), ref["URN"]
                ).codes

            else:
                short_urn = str(
                    Reference(
                        sdmx_type=ref[CLASS],
                        agency=ref[AGENCY_ID],
                        id=ref[ID],
                        version=ref[VERSION],
                    )
                )
                codelist = self.codelists[short_urn]

            json_obj[CODES_LOW] = codelist
        if ENUM_FORMAT in json_rep:
            self.__format_facets(json_rep[ENUM_FORMAT], json_obj)

    def __format_local_rep(self, representation_info: Dict[str, Any]) -> None:
        rep: Dict[str, Any] = {}

        self.__format_representation(representation_info[LOCAL_REP], rep)
        del representation_info[LOCAL_REP]

        if CODES_LOW in rep:
            representation_info[LOCAL_CODES_LOW] = rep.pop(CODES_LOW)

        if DTYPE in rep:
            representation_info[LOCAL_DTYPE] = rep.pop(DTYPE)

        if FACETS.lower() in rep:
            representation_info[LOCAL_FACETS_LOW] = rep.pop(FACETS.lower())

    def __format_con_id(self, concept_ref: Dict[str, Any]) -> Dict[str, Any]:
        rep = {}
        item_reference = ItemReference(
            sdmx_type=concept_ref[CLASS],
            agency=concept_ref[AGENCY_ID],
            id=concept_ref[PAR_ID],
            version=concept_ref[PAR_VER],
            item_id=concept_ref[ID],
        )
        scheme_reference = Reference(
            sdmx_type=CS,
            agency=concept_ref[AGENCY_ID],
            id=concept_ref[PAR_ID],
            version=concept_ref[PAR_VER],
        )

        concept_scheme = self.concepts.get(str(scheme_reference))
        if concept_scheme is None:
            return {CON: item_reference}
        for con in concept_scheme.concepts:
            if con.id == concept_ref[ID]:
                rep[CON] = con
                break
        if CON not in rep:
            return {CON: item_reference}
        return rep

    @staticmethod
    def __format_relationship(json_rel: Dict[str, Any]) -> Optional[str]:
        att_level = None

        for scheme in [DIM, PRIM_MEASURE]:
            if scheme in json_rel:
                if scheme == DIM:
                    dims = add_list(json_rel[DIM])
                    dims = [dim[REF][ID] for dim in dims]
                    att_level = ",".join(dims)
                else:
                    att_level = "O"

        return att_level

    def __format_vtl_references(
        self, json_elem: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Formats the references in the VTL element."""

        def extract_references(
            scheme: str, new_key: str, object_list: Dict[str, Any]
        ) -> None:
            if scheme in json_elem:
                references = []
                scheme_entries = (
                    json_elem[scheme]
                    if isinstance(json_elem[scheme], list)
                    else [json_elem[scheme]]
                )
                for entry in scheme_entries:
                    ref = entry[REF]

                    ref_id = ref[ID]
                    matching_object = next(
                        (
                            obj
                            for obj in object_list.values()
                            if getattr(obj, ID, None) == ref_id
                        ),
                        None,
                    )

                    if matching_object:
                        references.append(matching_object)
                    else:
                        references.append(
                            Reference(
                                sdmx_type=ref[CLASS],
                                agency=ref[AGENCY_ID],
                                id=ref_id,
                                version=ref[VERSION],
                            )
                        )

                json_elem[new_key] = references
                json_elem.pop(scheme)

        extract_references(RULE_SCHEME, "ruleset_schemes", self.rulesets)
        extract_references(
            UDO_SCHEME, "user_defined_operator_schemes", self.udos
        )

        return json_elem

    def __format_component(
        self, comp: Dict[str, Any], role: Role
    ) -> Component:
        comp[ROLE.lower()] = role
        comp[REQUIRED] = True

        self.__format_local_rep(comp) if LOCAL_REP in comp else None

        concept_id = self.__format_con_id(comp[CON_ID][REF])
        comp[CON_LOW] = concept_id.pop(CON)
        del comp[CON_ID]

        # Attribute Handling
        if ATT_REL in comp:
            comp[ATT_LVL] = self.__format_relationship(comp[ATT_REL])
            del comp[ATT_REL]

        if AS_STATUS in comp:
            if comp[AS_STATUS] != MANDATORY:
                comp[REQUIRED] = False
            del comp[AS_STATUS]

        if "position" in comp:
            del comp["position"]

        if ANNOTATIONS in comp:
            del comp[ANNOTATIONS]

        if CON_ROLE in comp:
            del comp[CON_ROLE]

        return Component(**comp)

    def __format_component_lists(
        self, element: Dict[str, Any]
    ) -> List[Component]:
        comp_list = []

        if TIME_DIM in element:
            element[DIM] = add_list(element[DIM])
            element[DIM].append(element[TIME_DIM])
            del element[TIME_DIM]

        role_name = list(set(element.keys()).intersection(COMP_TYPES))[0]
        role = ROLE_MAPPING[role_name]
        element[role_name] = add_list(element[role_name])

        for comp in element[role_name]:
            formatted_comp = self.__format_component(comp, role)
            comp_list.append(formatted_comp)

        return comp_list

    def __format_components(self, element: Dict[str, Any]) -> Dict[str, Any]:
        if DSD_COMPS in element:
            element[COMPS] = []
            comps = element[DSD_COMPS]

            for comp_list in [DIM_LIST, ME_LIST, GROUP, ATT_LIST]:
                if comp_list == GROUP and comp_list in comps:
                    del comps[GROUP]

                elif comp_list in comps:
                    fmt_comps = self.__format_component_lists(comps[comp_list])
                    element[COMPS].extend(fmt_comps)

            element[COMPS] = Components(element[COMPS])
            del element[DSD_COMPS]

        return element

    @staticmethod
    def __format_vtl(json_vtl: Dict[str, Any]) -> Dict[str, Any]:
        if "isPersistent" in json_vtl:
            json_vtl["is_persistent"] = json_vtl.pop("isPersistent")
        if "Expression" in json_vtl:
            json_vtl["expression"] = json_vtl.pop("Expression")
        if "Result" in json_vtl:
            json_vtl["result"] = json_vtl.pop("Result")
        if "vtlVersion" in json_vtl:
            json_vtl["vtl_version"] = json_vtl.pop("vtlVersion")
        if "rulesetScope" in json_vtl:
            json_vtl["ruleset_scope"] = json_vtl.pop("rulesetScope")
        if "rulesetType" in json_vtl:
            json_vtl["ruleset_type"] = json_vtl.pop("rulesetType")
        if "RulesetDefinition" in json_vtl:
            json_vtl["ruleset_definition"] = json_vtl.pop("RulesetDefinition")
        if "OperatorDefinition" in json_vtl:
            json_vtl["operator_definition"] = json_vtl.pop(
                "OperatorDefinition"
            )
        return json_vtl

    def __format_item(
        self, item_json_info: Dict[str, Any], item_name_class: str
    ) -> Item:
        item_json_info = self.__format_annotations(item_json_info)
        item_json_info = self.__format_name_description(item_json_info)
        if CONTACT in item_json_info and item_name_class == AGENCY:
            item_json_info[CONTACT] = add_list(item_json_info[CONTACT])
            contacts = []
            for e in item_json_info[CONTACT]:
                contacts.append(self.__format_contact(e))
            item_json_info[CONTACT.lower() + "s"] = contacts
            del item_json_info[CONTACT]

        if CORE_REP in item_json_info and item_name_class == CON:
            self.__format_representation(
                item_json_info[CORE_REP], item_json_info
            )
            del item_json_info[CORE_REP]

        if "Parent" in item_json_info:
            del item_json_info["Parent"]

        item_json_info = self.__format_vtl(item_json_info)

        return ITEMS_CLASSES[item_name_class](**item_json_info)

    def __format_scheme(
        self, json_elem: Dict[str, Any], scheme: str, item: str
    ) -> Dict[str, ItemScheme]:
        elements: Dict[str, ItemScheme] = {}

        json_elem[scheme] = add_list(json_elem[scheme])
        for element in json_elem[scheme]:
            element["items"] = []

            element = self.__format_annotations(element)
            element = self.__format_name_description(element)
            element = self.__format_urls(element)
            if IS_EXTERNAL_REF in element:
                element[IS_EXTERNAL_REF_LOW] = (
                    element.pop(IS_EXTERNAL_REF) == "true"
                )
            if IS_FINAL in element:
                element[IS_FINAL_LOW] = element.pop(IS_FINAL) == "true"
            if IS_PARTIAL in element:
                element[IS_PARTIAL_LOW] = element.pop(IS_PARTIAL) == "true"
            element[item] = add_list(element[item])
            items = []
            for item_elem in element[item]:
                # Dynamic
                items.append(self.__format_item(item_elem, item))
            del element[item]
            element["items"] = items
            element = self.__format_agency(element)
            element = self.__format_validity(element)
            element = self.__format_vtl(element)
            element = self.__format_vtl_references(element)
            if "xmlns" in element:
                del element["xmlns"]
            # Dynamic creation with specific class
            result: ItemScheme = STRUCTURES_MAPPING[scheme](**element)
            elements[result.short_urn] = result

        return elements

    def __format_schema(
        self, json_element: Dict[str, Any], schema: str, item: str
    ) -> Dict[str, Any]:
        """Formats the structures in json format.

        Args:
            json_element: The structures in json format
            schema: The scheme of the structures
            item: The item of the structures

        Returns:
            A dictionary with the structures formatted
        """
        schemas = {}

        json_element[item] = add_list(json_element[item])
        for element in json_element[item]:
            if URN.lower() in element and element[URN.lower()] is not None:
                short_urn = parse_urn(element[URN.lower()]).__str__()
            else:
                short_urn = Reference(
                    sdmx_type=item,
                    agency=element[AGENCY_ID],
                    id=element[ID],
                    version=element[VERSION],
                ).__str__()

            element = self.__format_annotations(element)
            element = self.__format_name_description(element)
            element = self.__format_urls(element)
            element = self.__format_agency(element)
            element = self.__format_validity(element)
            element = self.__format_components(element)

            if "xmlns" in element:
                del element["xmlns"]
            if IS_EXTERNAL_REF in element:
                element[IS_EXTERNAL_REF_LOW] = element.pop(IS_EXTERNAL_REF)
                element[IS_EXTERNAL_REF_LOW] = (
                    str(element[IS_EXTERNAL_REF_LOW]).lower() == "true"
                )
            if IS_FINAL in element:
                element[IS_FINAL_LOW] = element.pop(IS_FINAL)
                element[IS_FINAL_LOW] = (
                    str(element[IS_FINAL_LOW]).lower() == "true"
                )

            if item == DFW:
                ref_data = element[STRUCTURE][REF]
                reference_str = (
                    f"{ref_data[CLASS]}={ref_data[AGENCY_ID]}"
                    f":{ref_data[ID]}({ref_data[VERSION]})"
                )
                element[STRUCTURE] = reference_str

            structure = {key.lower(): value for key, value in element.items()}
            if schema == DSDS:
                if COMPS in structure:
                    structure[COMPS] = Components(structure[COMPS])
                else:
                    structure[COMPS] = Components([])
            schemas[short_urn] = STRUCTURES_MAPPING[schema](**structure)

        return schemas

    def format_structures(
        self, json_meta: Dict[str, Any]
    ) -> Sequence[Union[ItemScheme, DataStructureDefinition, Dataflow]]:
        """Formats the structures in JSON format.

        Args:
            json_meta: The structures in JSON format.

        Returns:
            A list with the formatted structures.
        """

        def process_structure(
            key: str,
            formatter: Callable[[Dict[str, Any]], Dict[Any, Any]],
            attr: Optional[str] = None,
        ) -> Dict[Any, Any]:
            """Helper function to process and store formatted structures."""
            if key in json_meta:
                formatted = formatter(json_meta[key])
                if attr:
                    setattr(self, attr, formatted)
                return formatted
            return {}

        structures = {
            ORGS: process_structure(ORGS, self.__format_orgs, "agencies"),
            CLS: process_structure(
                CLS,
                lambda data: self.__format_scheme(data, CL, CODE),
                "codelists",
            ),
            CONCEPTS: process_structure(
                CONCEPTS,
                lambda data: self.__format_scheme(data, CS, CON),
                "concepts",
            ),
            DSDS: process_structure(
                DSDS,
                lambda data: self.__format_schema(data, DSDS, DSD),
                "datastructures",
            ),
            DFWS: process_structure(
                DFWS, lambda data: self.__format_schema(data, DFWS, DFW)
            ),
            RULESETS: process_structure(
                RULESETS,
                lambda data: self.__format_scheme(data, RULE_SCHEME, RULE),
                "rulesets",
            ),
            UDOS: process_structure(
                UDOS,
                lambda data: self.__format_scheme(data, UDO_SCHEME, UDO),
                "udos",
            ),
            TRANSFORMATIONS: process_structure(
                TRANSFORMATIONS,
                lambda data: self.__format_scheme(
                    data,
                    TRANS_SCHEME,
                    TRANSFORMATION,
                ),
                "transformations",
            ),
        }
        return [
            compound
            for value in structures.values()
            if value
            for compound in value.values()
        ]


def read(
    input_str: str,
    validate: bool = True,
) -> Sequence[Union[ItemScheme, DataStructureDefinition, Dataflow]]:
    """Reads an SDMX-ML 2.1 Structure data and returns the structures.

    Args:
        input_str: SDMX-ML data to read.
        validate: If True, the XML data will be validated against the XSD.

    Returns:
        dict: Dictionary with the parsed structures.
    """
    dict_info = parse_xml(input_str, validate)
    if STRUCTURE not in dict_info:
        raise Invalid("This SDMX document is not SDMX-ML 2.1 Structure.")
    return StructureParser().format_structures(
        dict_info[STRUCTURE][STRUCTURES]
    )



"""Generates the full VTL Transformation Scheme script."""

from typing import Sequence, Union

from pysdmx.model import (
    Reference,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperatorScheme,
)
from pysdmx.toolkit.vtl.model_validations import model_validations


def _process_ruleset_scheme(
    ruleset_schemes: Sequence[Union[RulesetScheme, Reference]],
) -> str:
    """Processes the Ruleset Scheme objects and returns the VTL script."""
    vtl_script = ""
    for ruleset_scheme in ruleset_schemes:
        if isinstance(ruleset_scheme, RulesetScheme):
            for ruleset in ruleset_scheme.items:
                vtl_script += f"{ruleset.ruleset_definition}\n"

    return vtl_script


def _process_udo_scheme(
    udo_schemes: Sequence[Union[UserDefinedOperatorScheme, Reference]],
) -> str:
    """Processes the UDO Scheme objects and returns the VTL script."""
    vtl_script = ""
    for udo_scheme in udo_schemes:
        if isinstance(udo_scheme, UserDefinedOperatorScheme):
            for udo in udo_scheme.items:
                vtl_script += f"{udo.operator_definition}\n"

    return vtl_script


def _process_transformation(transformations: Sequence[Transformation]) -> str:
    """Processes the transformation objects and returns the VTL script."""
    vtl_script = ""
    for transformation in transformations:
        vtl_script += f"{transformation.full_expression}\n"

    return vtl_script


def generate_vtl_script(
    transformation_scheme: TransformationScheme, model_validation: bool = True
) -> str:
    """Generates the full VTL Transformation Scheme script.

    Args:
        transformation_scheme: A TransformationScheme object.
        model_validation: A boolean value to check if the model
                        object is valid.
                        if True, the model object is validated.

    returns:
        A string containing the full VTL Transformation Scheme script.

    """
    if model_validation:
        model_validations(transformation_scheme)

    vtl_script = ""

    vtl_script += _process_ruleset_scheme(
        transformation_scheme.ruleset_schemes
    )
    vtl_script += _process_udo_scheme(
        transformation_scheme.user_defined_operator_schemes
    )
    vtl_script += _process_transformation(transformation_scheme.items)

    return vtl_script



"""SDMX 2.1 XML StructureSpecificData reader module."""

import itertools
from typing import Any, Dict, Sequence

import numpy as np
import pandas as pd

from pysdmx.errors import Invalid
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.sdmx21.__tokens import (
    EXCLUDED_ATTRIBUTES,
    GROUP,
    OBS,
    SERIES,
    STR_REF,
    STR_SPE,
)
from pysdmx.io.xml.sdmx21.reader.__data_aux import (
    __process_df,
    get_data_objects,
)
from pysdmx.io.xml.sdmx21.reader.__parse_xml import parse_xml
from pysdmx.io.xml.utils import add_list
from pysdmx.model.dataset import ActionType


def __reading_str_series(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Structure Specific Series
    test_list = []
    df = None
    dataset[SERIES] = add_list(dataset[SERIES])
    for data in dataset[SERIES]:
        keys = dict(itertools.islice(data.items(), len(data)))
        if OBS in data:
            del keys[OBS]
            data[OBS] = add_list(data[OBS])
            for j in data[OBS]:
                test_list.append({**keys, **j})
        else:
            test_list.append(keys)
        test_list, df = __process_df(test_list, df)

    test_list, df = __process_df(test_list, df, is_end=True)

    return df


def __reading_group_data(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Structure Specific Group Data
    test_list = []
    df = None
    dataset[GROUP] = add_list(dataset[GROUP])
    for data in dataset[GROUP]:
        test_list.append(dict(data.items()))
        test_list, df = __process_df(test_list, df)
    test_list, df = __process_df(test_list, df, is_end=True)

    cols_to_delete = [x for x in df.columns if ":type" in x]
    for x in cols_to_delete:
        del df[x]

    df = df.drop_duplicates(keep="first").reset_index(drop=True)

    return df


def __get_at_att_str(dataset: Dict[str, Any]) -> Dict[str, Any]:
    """Gets the elements of the dataset if it is Structure Specific Data."""
    return {k: dataset[k] for k in dataset if k not in EXCLUDED_ATTRIBUTES}


def __parse_structure_specific_data(
    dataset: Dict[str, Any], structure_info: Dict[str, Any]
) -> PandasDataset:
    attached_attributes = __get_at_att_str(dataset)

    df = pd.DataFrame()

    # Parsing data
    if SERIES in dataset:
        # Structure Specific Series
        df = __reading_str_series(dataset)
        if GROUP in dataset:
            df_group = __reading_group_data(dataset)
            common_columns = list(
                set(df.columns).intersection(set(df_group.columns))
            )
            df = pd.merge(df, df_group, on=common_columns, how="left")
    elif OBS in dataset:
        dataset[OBS] = add_list(dataset[OBS])
        # Structure Specific All dimensions
        df = pd.DataFrame(dataset[OBS]).replace(np.nan, "")

    urn = f"{structure_info['structure_type']}={structure_info['unique_id']}"
    action = dataset.get("action", "Information")
    action = ActionType(action)

    return PandasDataset(
        structure=urn, attributes=attached_attributes, data=df, action=action
    )


def read(input_str: str, validate: bool = True) -> Sequence[PandasDataset]:
    """Reads an SDMX-ML 2.1 Generic file and returns a Sequence of Datasets.

    Args:
        input_str: SDMX-ML data to read.
        validate: If True, the XML data will be validated against the XSD.
    """
    dict_info = parse_xml(input_str, validate=validate)
    if STR_SPE not in dict_info:
        raise Invalid(
            "This SDMX document is not an SDMX-ML 2.1 StructureSpecificData."
        )
    dataset_info, str_info = get_data_objects(dict_info[STR_SPE])
    datasets = []
    for dataset in dataset_info:
        ds = __parse_structure_specific_data(
            dataset, str_info[dataset[STR_REF]]
        )
        datasets.append(ds)
    return datasets



"""Build SDMX-REST registration queries."""

from datetime import datetime
from typing import Optional, Sequence, Union

import msgspec
from msgspec.json import Decoder

from pysdmx.api.qb.data import DataContext
from pysdmx.api.qb.util import REST_ALL, ApiVersion, CoreQuery
from pysdmx.errors import Invalid
from pysdmx.io.format import RegistryFormat


class _CoreRegistrationQuery(
    CoreQuery,
    frozen=True,
    omit_defaults=True,
):
    def _check_version(self, version: ApiVersion) -> None:
        if version < ApiVersion.V2_1_0:
            raise Invalid(
                "Invalid Request",
                (
                    "Registration are not supported"
                    f"in SDMX-REST {version.value}."
                ),
            )

    def _check_updated_consistency(
        self,
        updated_before: Optional[datetime],
        updated_after: Optional[datetime],
    ) -> None:
        if updated_before and updated_after and updated_before < updated_after:
            raise Invalid(
                "Inconsistent updated timestamps",
                (
                    "The updated_after timestamp should be before "
                    "the updated_before timestamp."
                ),
                csi={
                    "updated_after": str(updated_after),
                    "updated_before": str(updated_before),
                },
            )

    def _create_qs(
        self,
        updb: Optional[datetime],
        upda: Optional[datetime],
    ) -> str:
        qs = super()._append_qs_param(
            "",
            updb,
            "updatedBefore",
            updb.isoformat("T", "seconds") if updb else None,
        )
        qs = super()._append_qs_param(
            qs,
            upda,
            "updatedAfter",
            upda.isoformat("T", "seconds") if upda else None,
        )
        if qs:
            qs = f"?{qs}"
        return qs


class RegistrationByIdQuery(
    _CoreRegistrationQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for registration, using the registraton ID.

    Attributes:
        registration_id: The registration id(s).
    """

    registration_id: str

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_id_decoder

    def _create_full_query(self, ver: ApiVersion) -> str:
        return f"/registration/id/{self.registration_id}"

    def _create_short_query(self, ver: ApiVersion) -> str:
        return self._create_full_query(ver)


class RegistrationByProviderQuery(
    _CoreRegistrationQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for registrations by provider.

    Attributes:
        provider_agency_id: The ID of the agency maintaining the data
            provider scheme.
        provider_id: The provider of the registered data.
        updated_before: Returns only registrations updated or created before
            the supplied timestamp.
        updated_after: Returns only registrations updated or created after
            the supplied timestamp.
    """

    provider_agency_id: Union[str, Sequence[str]] = REST_ALL
    provider_id: Union[str, Sequence[str]] = REST_ALL
    updated_before: Optional[datetime] = None
    updated_after: Optional[datetime] = None

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)
        super()._check_updated_consistency(
            self.updated_before, self.updated_after
        )

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_prov_decoder

    def _create_full_query(self, ver: ApiVersion) -> str:
        a = super()._join_mult(self.provider_agency_id)
        p = super()._join_mult(self.provider_id)
        q = super()._create_qs(self.updated_before, self.updated_after)
        return f"/registration/provider/{a}/{p}{q}"

    def _create_short_query(self, ver: ApiVersion) -> str:
        q = super()._create_qs(self.updated_before, self.updated_after)
        p = (
            f"/{super()._join_mult(self.provider_id)}"
            if self.provider_id != REST_ALL
            else ""
        )
        a = (
            f"/{super()._join_mult(self.provider_agency_id)}"
            if self.provider_agency_id != REST_ALL
            else ""
        )
        return f"/registration/provider{a}{p}{q}"


class RegistrationByContextQuery(
    _CoreRegistrationQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for registrations by context.

    Attributes:
        provider_agency_id: The ID of the agency maintaining the data
            provider scheme.
        provider_id: The provider of the registered data.
        updated_before: Returns only registrations updated or created before
            the supplied timestamp.
        updated_after: Returns only registrations updated or created after
            the supplied timestamp.
    """

    context: DataContext = DataContext.ALL
    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_ALL
    updated_before: Optional[datetime] = None
    updated_after: Optional[datetime] = None

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)
        super()._check_updated_consistency(
            self.updated_before, self.updated_after
        )

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_ctx_decoder

    def _create_full_query(self, ver: ApiVersion) -> str:
        o = f"/{self.context.value}"
        a = super()._join_mult(self.agency_id)
        r = super()._join_mult(self.resource_id)
        v = super()._join_mult(self.version)
        o += f"/{a}/{r}/{v}"
        q = super()._create_qs(self.updated_before, self.updated_after)
        return f"/registration{o}{q}"

    def _create_short_query(self, ver: ApiVersion) -> str:
        v = (
            f"/{super()._join_mult(self.version)}"
            if self.version != REST_ALL
            else ""
        )
        r = (
            f"/{super()._join_mult(self.resource_id)}{v}"
            if v or self.resource_id != REST_ALL
            else ""
        )
        a = (
            f"/{super()._join_mult(self.agency_id)}{r}"
            if r or self.agency_id != REST_ALL
            else ""
        )
        c = (
            f"/{self.context.value}{a}"
            if a or self.context != DataContext.ALL
            else ""
        )
        q = super()._create_qs(self.updated_before, self.updated_after)
        return f"/registration{c}{q}"


_by_id_decoder = msgspec.json.Decoder(RegistrationByIdQuery)
_by_ctx_decoder = msgspec.json.Decoder(RegistrationByContextQuery)
_by_prov_decoder = msgspec.json.Decoder(RegistrationByProviderQuery)

__all__ = [
    "RegistrationByIdQuery",
    "RegistrationByProviderQuery",
    "RegistrationByContextQuery",
    "RegistryFormat",
]



"""Module for writing SDMX-ML 2.1 Error messages."""

from pysdmx.io.format import Format
from pysdmx.io.xml.sdmx21.writer.__write_aux import __namespaces_from_type


def write() -> None:
    """Write data to SDMX-ML 2.1 Error format.

    Returns:
        The XML string if path is empty, None otherwise.
    """
    __namespaces_from_type(Format.ERROR_SDMX_ML_2_1)



"""Schemas for Fusion-JSON messages."""

from pysdmx.io.json.fusion.messages.category import (
    FusionCategorisationMessage,
    FusionCategorySchemeMessage,
)
from pysdmx.io.json.fusion.messages.code import (
    FusionCodelistMessage,
    FusionHierarchyAssociationMessage,
    FusionHierarchyMessage,
)
from pysdmx.io.json.fusion.messages.concept import FusionConceptSchemeMessage
from pysdmx.io.json.fusion.messages.dataflow import (
    FusionDataflowMessage,
    FusionDataflowsMessage,
)
from pysdmx.io.json.fusion.messages.map import (
    FusionMappingMessage,
    FusionRepresentationMapMessage,
)
from pysdmx.io.json.fusion.messages.org import (
    FusionAgencyMessage,
    FusionProviderMessage,
)
from pysdmx.io.json.fusion.messages.pa import (
    FusionProvisionAgreementMessage as FusionPAMessage,
)
from pysdmx.io.json.fusion.messages.report import FusionMetadataMessage
from pysdmx.io.json.fusion.messages.schema import FusionSchemaMessage
from pysdmx.io.json.fusion.messages.vtl import (
    FusionVtlTransformationsMessage as FusionTransfoMsg,
)

__all__ = [
    "FusionCategorisationMessage",
    "FusionCategorySchemeMessage",
    "FusionCodelistMessage",
    "FusionHierarchyAssociationMessage",
    "FusionHierarchyMessage",
    "FusionConceptSchemeMessage",
    "FusionDataflowMessage",
    "FusionDataflowsMessage",
    "FusionMappingMessage",
    "FusionRepresentationMapMessage",
    "FusionAgencyMessage",
    "FusionProviderMessage",
    "FusionMetadataMessage",
    "FusionSchemaMessage",
    "FusionTransfoMsg",
    "FusionPAMessage",
]
